"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_output_headers, get_resistance_results


class TestResTem(unittest.TestCase):
    """
    This test is for a particular bug we found, where BLAST can find an exact amino acid match but
    on the wrong strand. This test sequence was coming up as "TEM-15^" from a wrong strand match
    until we fixed the bug (only checking the forward strand).
    """
    def setUp(self):
        self.data_dir = 'test/test_res_tem/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def test_tem(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_tem/tem.fasta',
                                         self.args, self.res_headers, True)
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'TEM-1D^')



#!/usr/bin/env python3

"""
This script generates a table of beta-lactamase class information.

It takes two arguments:
  1) The Lahey info file at ftp://ftp.ncbi.nlm.nih.gov/pathogen/betalactamases/Lahey.tab
  2) A GenBank file downloaded from these records:
     https://www.ncbi.nlm.nih.gov/nuccore?term=313047%5BBioProject%5D
     (use 'Send to' then 'File' and set the format to GenBank)

Usage:
  ./bla_info.py Lahey.tab sequence.gb > bla_info_table

The output has three columns:
  * allele name
  * allele description
  * bla class

The bla class column can have these six possible values:
  * Bla_broad (broad spectrum beta-lactamases)
  * Bla_broad_inhR (broad spectrum beta-lactamases with resistance to beta-lactamase inhibitors)
  * Bla_ESBL (extended spectrum beta-lactamases)
  * Bla_ESBL_inhR (extended spectrum beta-lactamases with resistance to beta-lactamase inhibitors)
  * Bla_Carb (carbapenemase)
  * Bla (beta-lactamase that is none of the above)
"""

import sys
import re
from Bio import SeqIO


def main():
    lahey_filename = sys.argv[1]
    genbank_filename = sys.argv[2]

    lahey_descriptions = get_lahey_descriptions(lahey_filename)
    genbank_descriptions = get_genbank_descriptions(genbank_filename)

    all_alleles = sorted(set(list(lahey_descriptions.keys()) + list(genbank_descriptions.keys())),
                         key=allele_name_sorting_key)

    for allele in all_alleles:

        # If the allele is in only one of the two sources, then we just get the description/class
        # from the single source.
        if allele in lahey_descriptions and allele not in genbank_descriptions:
            description = lahey_descriptions[allele]
            bla_class = bla_class_from_description(description)
        elif allele in genbank_descriptions and allele not in lahey_descriptions:
            description = genbank_descriptions[allele]
            bla_class = bla_class_from_description(description)

        # If the allele is in both Lahey and the GenBank...
        else:
            lahey_description = lahey_descriptions[allele]
            genbank_description = genbank_descriptions[allele]
            lahey_bla_class = bla_class_from_description(lahey_description)
            genbank_bla_class = bla_class_from_description(genbank_description)

            # If they agree on the class, then we use the longer description (assuming that the
            # longer one is more descriptive).
            if lahey_bla_class == genbank_bla_class:
                bla_class = lahey_bla_class
                description = lahey_description
                if len(genbank_description) > len(lahey_description):
                    description = genbank_description

            # If they disagree on the class, then we choose whichever class is 'worse'.
            else:
                class_values  = {'Bla': 0,
                                 'Bla_broad': 1, 'Bla_broad_inhR': 2,
                                 'Bla_ESBL': 3, 'Bla_ESBL_inhR': 4,
                                 'Bla_Carb': 5}
                if class_values[lahey_bla_class] > class_values[genbank_bla_class]:
                    bla_class = lahey_bla_class
                    description = lahey_description
                else:
                    bla_class = genbank_bla_class
                    description = genbank_description

        print('\t'.join([allele, description, bla_class]))


def get_lahey_descriptions(lahey_filename):
    descriptions = {}
    with open(lahey_filename, 'rt') as lahey_file:
        for line in lahey_file:
            if line.startswith('#'):
                continue
            line_parts = line.strip().split('\t')
            allele = line_parts[0]
            descriptions[allele] = line_parts[7].rsplit(' ', 1)[0]
    return descriptions


def get_genbank_descriptions(genbank_filename):
    descriptions = {}
    for seq_record in SeqIO.parse(sys.argv[2], 'genbank'):
        for feature in seq_record.features:
            if feature.type == 'CDS':
                allele = feature.qualifiers.get('allele', ['none'])[0]
                gene = feature.qualifiers.get('gene', ['none'])[0]
                product = feature.qualifiers.get('product', [''])[0]
                note = feature.qualifiers.get('note', [''])[0]
                function = feature.qualifiers.get('function', [''])[0]

                if allele.lower().startswith('bla') and gene.lower().startswith('bla'):
                    allele = allele[3:]  # remove 'bla'
                    if allele not in descriptions:  # Don't overwrite the Lahey description.

                        # The product likely ends with the allele name. If so, remove it.
                        allele_name_len = len(allele)
                        if allele != 'none' and allele_name_len and \
                                product.lower().endswith(allele.lower()):
                            product = product[:-allele_name_len]
                        product = product.rstrip()
                        if product.lower().endswith(' bla'):
                            product = product[:-4]

                        # The 'product' feature probably has all of the description we need.
                        # However, if there is additional information in the 'note' or 'function'
                        # features, then add those too.
                        description = product
                        if ('carbapenem' in note and 'carbapenem' not in description) or \
                                ('extended' in note and 'extended' not in description) or \
                                ('broad' in note and 'broad' not in description) or \
                                ('inhibitor' in note and 'inhibitor' not in description):
                            description += ' ' + note
                        if ('carbapenem' in function and 'carbapenem' not in description) or \
                                ('extended' in function and 'extended' not in description) or \
                                ('broad' in function and 'broad' not in description) or \
                                ('inhibitor' in function and 'inhibitor' not in description):
                            description += ' ' + function
                        description = description.replace(',', ' ')
                        description = description.replace(';', ' ')
                        description = re.sub('\s+', ' ', description).strip()

                        # Special case to fix a particular long description.
                        if description == 'subclass B1 metallo-beta-lactamase New Delhi ' + \
                                          'metallo-beta-lactamase+ New Delhi MBL carbapenemase':
                            description = 'subclass B1 metallo-beta-lactamase carbapenemase'

                        descriptions[allele] = description
    return descriptions


def bla_class_from_description(description):
    if 'carbapenem-hydrolyzing' in description or \
            'carbapenemase' in description or \
            'confers carbapenem resistance' in description or \
            'hydrolyses beta-lactams including carbapenems' in description:
        return 'Bla_Carb'
    elif 'extended-spectrum' in description or \
            'extended spectrum' in description or \
            'ESBL' in description:
        if 'inhibitor-resistant' in description:
            return 'Bla_ESBL_inhR'
        else:
            return 'Bla_ESBL'
    elif 'broad-spectrum' in description or \
            'broad spectrum' in description:
        if 'inhibitor-resistant' in description:
            return 'Bla_broad_inhR'
        else:
            return 'Bla_broad'
    return 'Bla'


def allele_name_sorting_key(allele_name):
    """
    This function defines a sorting key for allele names. It puts padding zeros in front of
    numbers so 'TEM-2' will come before 'TEM-10', etc.
    """
    new_name_parts = []
    for name_part in allele_name.split('-'):
        try:
            name_part_int = int(name_part)
            new_name_parts.append('%05d' % name_part_int)
        except ValueError:
            new_name_parts.append(name_part)
    return '-'.join(new_name_parts)


if __name__ == '__main__':
    main()



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_output_headers, get_resistance_results


class TestResOmp(unittest.TestCase):
    """
    Tests calling of carbapenem resistance via the OmpK35/OmpK36 genes.
    """

    def setUp(self):
        self.data_dir = 'test/test_res_omp/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def test_both_genes_intact(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_1.fasta',
                                         self.args, self.res_headers, True)
        self.assertEqual(results['Omp_mutations'], '-')

    def test_ompk35_frameshift(self):
        """
        A frameshift in OmpK35 should cause an early stop and lead to a carbapenem resistance call.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_2.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('OmpK35-' in results['Omp_mutations'])

    def test_ompk35_early_stop(self):
        """
        This tests an early stop mutation (without a frameshift) in OmpK35.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_3.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('OmpK35-' in results['Omp_mutations'])

    def test_ompk36_missing(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_4.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('OmpK36-' in results['Omp_mutations'])

    def test_ompk36gd(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_5.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('OmpK36GD' in results['Omp_mutations'])

    def test_ompk36td(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_6.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('OmpK36TD' in results['Omp_mutations'])

    def test_ompk36td_non_kp_complex(self):
        """
        Setting the Kp complex variable to False should turn off the OmpK tests.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_omp/test_res_omp_6.fasta',
                                         self.args, self.res_headers, False)
        self.assertTrue('OmpK36TD' not in results['Omp_mutations'])



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import distutils.spawn
import gzip
import os
import pathlib
import subprocess
import sys
import tempfile

from pkg_resources import resource_filename
from .contig_stats import get_contig_stat_results
from .help_formatter import MyParser, MyHelpFormatter
from .kaptive import get_kaptive_paths, get_kaptive_results
from .species import get_species_results, is_kp_complex
from .mlstBLAST import mlst_blast
from .resBLAST import read_class_file, get_res_headers, resblast_one_assembly
from .rmpA import rmpa2_blast
from .misc import get_compression_type, load_fasta
from .version import __version__


def main():
    args = parse_arguments()
    check_inputs_and_programs(args)
    data_folder = get_data_path()
    if args.force_index:
        rebuild_blast_indices(data_folder)
    kaptive_py, kaptive_k_db, kaptive_o_db = get_kaptive_paths()

    stdout_header, full_header, res_headers = get_output_headers(args, data_folder)
    output_headers(stdout_header, full_header, args.outfile)

    for contigs in args.assemblies:
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)

            # All results are stored in a dictionary where the key is the column name and the value
            # is the result. The results are outputted in order of the header rows. This means that
            # the column orders can be easily changed by modifying the get_output_headers function.

            results = {'strain': get_strain_name(contigs)}
            results.update(get_species_results(contigs, data_folder))
            kp_complex = is_kp_complex(results)
            results.update(get_contig_stat_results(contigs, kp_complex))

            results.update(get_chromosome_mlst_results(data_folder, contigs, kp_complex, args))
            results.update(get_all_virulence_results(data_folder, contigs, args))
            results.update(get_rmpa2_results(data_folder, contigs, args))
            results.update(get_wzi_and_k_locus_results(data_folder, contigs, args))
            results.update(get_resistance_results(data_folder, contigs, args, res_headers,
                                                  kp_complex))
            results.update(get_summary_results(results, res_headers, args))
            results.update(get_kaptive_results('K', kaptive_py, kaptive_k_db, contigs, args))
            results.update(get_kaptive_results('O', kaptive_py, kaptive_o_db, contigs, args))

            output_results(stdout_header, full_header, args.outfile, results)


def parse_arguments():
    parser = MyParser(description='Kleborate: a tool for characterising virulence and resistance '
                                  'in Klebsiella',
                      formatter_class=MyHelpFormatter, add_help=False,
                      epilog='R|If you use Kleborate, please cite the paper:\n'
                             '    Lam MMC, et al. A genomic surveillance framework and genotyping tool\n'
                             '    for Klebsiella pneumoniae and its related species complex. Nature\n'
                             '    Communications. 2021. doi:10.1038/s41467-021-24448-3.\n\n'
                             'If you turn on the Kaptive option for full K and O typing, please also cite Kaptive:\n'
                             '    Wyres KL, et al. Identification of Klebsiella capsule synthesis loci from\n'
                             '    whole genome data. Microbial Genomics. 2016. doi:10.1099/mgen.0.000102.')

    required_args = parser.add_argument_group('Required arguments')
    required_args.add_argument('-a', '--assemblies', nargs='+', type=str, required=True,
                               help='FASTA file(s) for assemblies')

    screening_args = parser.add_argument_group('Screening options')
    screening_args.add_argument('-r', '--resistance', action='store_true',
                                help='Turn on resistance genes screening (default: no resistance '
                                     'gene screening)')
    screening_args.add_argument('--kaptive_k', action='store_true',
                                help='Turn on Kaptive screening of K loci (default: do not run '
                                     'Kaptive for K loci)')
    screening_args.add_argument('--kaptive_o', action='store_true',
                                help='Turn on Kaptive screening of O loci (default: do not run '
                                     'Kaptive for O loci)')
    screening_args.add_argument('-k', '--kaptive', action='store_true',
                                help='Equivalent to --kaptive_k --kaptive_o NO_DEFAULT')
    screening_args.add_argument('--all', action='store_true',
                                help='Equivalent to --resistance --kaptive NO_DEFAULT')

    output_args = parser.add_argument_group('Output options')
    output_args.add_argument('-o', '--outfile', type=str, default='Kleborate_results.txt',
                             help='File for detailed output (default: Kleborate_results.txt)')
    output_args.add_argument('--kaptive_k_outfile', type=str,
                             help='File for full Kaptive K locus output (default: do not '
                                  'save Kaptive K locus results to separate file)')
    output_args.add_argument('--kaptive_o_outfile', type=str,
                             help='File for full Kaptive O locus output (default: do not '
                                  'save Kaptive O locus results to separate file)')

    setting_args = parser.add_argument_group('Settings')
    setting_args.add_argument('--min_identity', type=float, default=90.0,
                              help='Minimum alignment percent identity for main results')
    setting_args.add_argument('--min_coverage', type=float, default=80.0,
                              help='Minimum alignment percent coverage for main results')
    setting_args.add_argument('--min_spurious_identity', type=float, default=80.0,
                              help='Minimum alignment percent identity for spurious results')
    setting_args.add_argument('--min_spurious_coverage', type=float, default=40.0,
                              help='Minimum alignment percent coverage for spurious results')
    setting_args.add_argument('--min_kaptive_confidence', type=str,
                              choices=['None', 'Low', 'Good', 'High', 'Very_high', 'Perfect'],
                              default='Good',
                              help='Minimum Kaptive confidence to call K/O loci - confidence '
                                   'levels below this will be reported as unknown')

    other_args = parser.add_argument_group('Other')
    other_args.add_argument('--force_index', action='store_true',
                            help='Rebuild the BLAST index at the start of execution (default: '
                                 'only build BLAST indices when they are missing)')

    help_args = parser.add_argument_group('Help')
    help_args.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                           help='Show this help message and exit')
    help_args.add_argument('--version', action='version', version='Kleborate v' + __version__,
                           help="Show program's version number and exit")

    # If no arguments were used, print the entire help (argparse default is to just give an error
    # like '-a is required').
    if len(sys.argv) == 1:
        parser.print_help(file=sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if args.kaptive:
        args.kaptive_k = True
        args.kaptive_o = True
    if args.all:
        args.resistance = True
        args.kaptive_k = True
        args.kaptive_o = True

    if args.kaptive_k_outfile and not args.kaptive_k:
        sys.exit('Error: you must use --kaptive_k (or --kaptive) to use --kaptive_k_outfile')
    if args.kaptive_o_outfile and not args.kaptive_o:
        sys.exit('Error: you must use --kaptive_o (or --kaptive) to use --kaptive_o_outfile')

    return args


def check_inputs_and_programs(args):
    for assembly in args.assemblies:
        if os.path.isdir(assembly):
            sys.exit('Error: ' + assembly + ' is a directory (please specify assembly files)')
        if not os.path.isfile(assembly):
            sys.exit('Error: could not find ' + assembly)
        fasta = load_fasta(assembly)
        if len(fasta) < 1:
            sys.exit('Error: invalid FASTA file: ' + assembly)
        for record in fasta:
            header, seq = record
            if len(seq) == 0:
                sys.exit('Error: invalid FASTA file (contains a zero-length sequence): ' + assembly)
    if not distutils.spawn.find_executable('makeblastdb'):
        sys.exit('Error: could not find makeblastdb')
    if not distutils.spawn.find_executable('blastn'):
        sys.exit('Error: could not find blastn')
    if args.resistance:
        if not distutils.spawn.find_executable('blastx'):
            sys.exit('Error: could not find blastx')
    if not distutils.spawn.find_executable('mash'):
        sys.exit('Error: could not find mash')
    major, minor, patch = get_blast_version()
    if major < 2 or (major == 2 and minor < 7):
        sys.exit('Error: you have BLAST v{}.{}.{} installed, but Kleborate requires v2.7.1 or '
                 'later'.format(major, minor, patch))


def get_blast_version():
    command = ['blastn', '-version']
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, _ = process.communicate()
    out = out.decode()
    try:
        version = out.split(': ')[1].split()[0].split('+')[0]
        major, minor, patch = version.split('.')
        return int(major), int(minor), patch
    except (IndexError, ValueError):
        sys.exit('Error: could not determine BLAST version')


def get_output_headers(args, data_folder):
    """
    There are two levels of output:
      * stdout is simpler and displayed to the console
      * full contains more and is saved to file
    This function returns headers for both. It also returns the resistance headers in a separate
    list, as they are used to total up some resistance summaries.
    """
    stdout_header = ['strain', 'species']
    full_header = ['strain', 'species', 'species_match']
    stdout_header += ['ST', 'virulence_score']
    full_header += ['contig_count', 'N50', 'largest_contig', 'total_size', 'ambiguous_bases',
                    'QC_warnings', 'ST', 'virulence_score']

    if args.resistance:
        stdout_header.append('resistance_score')
        full_header.append('resistance_score')
        full_header.append('num_resistance_classes')
        full_header.append('num_resistance_genes')

    other_columns = ['Yersiniabactin', 'YbST',
                     'Colibactin', 'CbST',
                     'Aerobactin', 'AbST',
                     'Salmochelin', 'SmST',
                     'RmpADC', 'RmST',
                     'rmpA2',
                     'wzi', 'K_locus']
    stdout_header += other_columns
    full_header += other_columns

    if args.kaptive_k:
        stdout_header.append('K_locus_confidence')
        full_header.append('K_type')
        full_header.append('K_locus_problems')
        full_header.append('K_locus_confidence')
        full_header.append('K_locus_identity')
        full_header.append('K_locus_missing_genes')

    if args.kaptive_o:
        stdout_header.append('O_locus')
        stdout_header.append('O_locus_confidence')
        full_header.append('O_locus')
        full_header.append('O_type')
        full_header.append('O_locus_problems')
        full_header.append('O_locus_confidence')
        full_header.append('O_locus_identity')
        full_header.append('O_locus_missing_genes')

    if args.resistance:
        gene_info, res_classes, bla_classes = \
            read_class_file(data_folder + '/CARD_AMR_clustered.csv')
        res_headers = get_res_headers(res_classes, bla_classes)
        res_headers += ['truncated_resistance_hits', 'spurious_resistance_hits']
        stdout_header += res_headers
        full_header += res_headers
    else:
        res_headers = []

    full_header.append('Chr_ST')
    full_header += get_chromosome_mlst_header()
    full_header += get_ybt_mlst_header()
    full_header += get_clb_mlst_header()
    full_header += get_iuc_mlst_header()
    full_header += get_iro_mlst_header()
    full_header += get_rmp_mlst_header()

    full_header.append('spurious_virulence_hits')

    return stdout_header, full_header, res_headers


def get_virulence_score(yersiniabactin, colibactin, aerobactin):
    """
    Six possible virulence scores:
      * 0 = no virulence (no yersiniabactin, colibactin or aerobactin)
      * 1 = just yersiniabactin (no colibactin or aerobactin)
      * 2 = colibactin but no aerobactin (regardless of yersiniabactin, which is probably present)
      * 3 = just aerobactin (no yersiniabactin or colibactin)
      * 4 = aerobactin and yersiniabactin (but not colibactin)
      * 5 = colibactin and aerobactin (regardless of yersiniabactin, which is probably present)
    """
    has_ybt = (yersiniabactin != '-')
    has_aero = (aerobactin != '-')
    has_coli = (colibactin != '-')

    if has_coli and has_aero:
        return 5
    elif has_aero and has_ybt:
        return 4
    elif has_aero:
        return 3
    elif has_coli:
        return 2
    elif has_ybt:
        return 1
    else:
        return 0


def get_resistance_score(res_headers, res_hits):
    """
    Four possible resistance scores:
      * 0 = no ESBL, no carbapenemase (regardless of colistin resistance)
      * 1 = ESBL, no carbapenemase (regardless of colistin resistance)
      * 2 = Carbapenemase without colistin resistance
      * 3 = Carbapenemase and colistin resistance
    """
    if not res_headers:
        return '-'

    # Look for a hit in any 'ESBL' column (e.g. 'Bla_ESBL' or 'Bla_ESBL_inhR').
    esbl_header_indices = [i for i, h in enumerate(res_headers) if '_esbl' in h.lower()]
    has_esbl = any(res_hits[i] != '-' for i in esbl_header_indices)

    # Look for a hit in any 'Carb' column (e.g. 'Bla_Carb').
    carb_header_indices = [i for i, h in enumerate(res_headers) if '_carb' in h.lower()]
    has_carb = any(res_hits[i] != '-' for i in carb_header_indices)

    # Look for a hit in the 'Col' columns.
    col_header_indices = [i for i, h in enumerate(res_headers)
                          if h.lower() == 'col_acquired' or h.lower() == 'col_mutations']
    has_col = any(res_hits[i] != '-' for i in col_header_indices)

    if has_carb and has_col:
        return 3
    elif has_carb:
        return 2
    elif has_esbl:
        return 1
    else:
        return 0


def get_resistance_class_count(res_headers, res_hits):
    """
    Counts up all resistance gene classes, excluding the 'Bla_chr' class which is intrinsic.
    """
    if not res_headers:
        return '-'
    res_classes = [h for i, h in enumerate(res_headers) if
                   res_hits[i] != '-' and
                   (h.lower().endswith('_acquired') or h == 'Col_mutations' or
                    h == 'Flq_mutations')]
    res_classes = [c.replace('_acquired', '').replace('_mutations', '') for c in res_classes]
    return len(set(res_classes))


def get_resistance_gene_count(res_headers, res_hits):
    """
    Counts up all resistance genes, excluding the 'Bla' class which is intrinsic.
    """
    if not res_headers:
        return '-'
    res_indices = [i for i, h in enumerate(res_headers) if h.lower().endswith('_acquired')]
    gene_list = []
    for i in res_indices:
        genes = res_hits[i].split(';')
        genes = [g for g in genes if g != '-']
        gene_list += genes
    return len(gene_list)


def get_data_path():
    return resource_filename(__name__, 'data')


def get_chromosome_mlst_header():
    return ['gapA', 'infB', 'mdh', 'pgi', 'phoE', 'rpoB', 'tonB']


def get_ybt_mlst_header():
    return ['ybtS', 'ybtX', 'ybtQ', 'ybtP', 'ybtA', 'irp2', 'irp1', 'ybtU', 'ybtT', 'ybtE', 'fyuA']


def get_clb_mlst_header():
    return ['clbA', 'clbB', 'clbC', 'clbD', 'clbE', 'clbF', 'clbG', 'clbH', 'clbI', 'clbL', 'clbM',
            'clbN', 'clbO', 'clbP', 'clbQ']


def get_iuc_mlst_header():
    return ['iucA', 'iucB', 'iucC', 'iucD', 'iutA']


def get_iro_mlst_header():
    return ['iroB', 'iroC', 'iroD', 'iroN']


def get_rmp_mlst_header():
    return ['rmpA', 'rmpD', 'rmpC']


def gunzip_contigs_if_necessary(contigs, temp_dir):
    if get_compression_type(contigs) == 'gz':
        name = get_strain_name(contigs)
        new_contigs = temp_dir + '/' + name + '.fasta'
        decompress_file(contigs, new_contigs)
        return new_contigs
    else:
        return contigs


def decompress_file(in_file, out_file):
    with gzip.GzipFile(in_file, 'rb') as i, open(out_file, 'wb') as o:
        s = i.read()
        o.write(s)


def get_chromosome_mlst_results(data_folder, contigs, kp_complex, args):
    chromosome_mlst_header = get_chromosome_mlst_header()

    if kp_complex:
        seqs = data_folder + '/Klebsiella_pneumoniae.fasta'
        database = data_folder + '/kpneumoniae.txt'
        chr_st, chr_st_detail, _, _ = \
            mlst_blast(seqs, database, 'no', [contigs], min_cov=args.min_coverage,
                       min_ident=args.min_identity, max_missing=3, allow_multiple=False)
        if chr_st != '0':
            chr_st = 'ST' + chr_st
        chr_st_with_subsp = get_kp_subspecies_based_on_st(chr_st)

        assert len(chromosome_mlst_header) == len(chr_st_detail)

        results = {'ST': chr_st_with_subsp,
                   'Chr_ST': chr_st}

    else:
        results = {'ST': "NA",
                   'Chr_ST': "NA"}
        chr_st_detail = ['-'] * len(chromosome_mlst_header)

    results.update(dict(zip(get_chromosome_mlst_header(), chr_st_detail)))
    return results


def get_kp_subspecies_based_on_st(chr_st):
    ozaenae_sts = {'ST90', 'ST91', 'ST92', 'ST93', 'ST95', 'ST96', 'ST97', 'ST381', 'ST777',
                   'ST3193', 'ST3766', 'ST3768', 'ST3771', 'ST3781', 'ST3782', 'ST3784', 'ST3802',
                   'ST3803'}
    rhinoscleromatis_sts = {'ST67', 'ST68', 'ST69', 'ST3772', 'ST3819'}
    chr_st_minus_1lv = chr_st.replace('-1LV', '')  # 1LV still results in a subspecies call
    if chr_st_minus_1lv in ozaenae_sts:
        return chr_st + ' (subsp. ozaenae)'
    if chr_st_minus_1lv in rhinoscleromatis_sts:
        return chr_st + ' (subsp. rhinoscleromatis)'
    return chr_st


def get_virulence_cluster_results(data_folder, contigs, alleles_fasta, profiles_txt,
                                  vir_name, vir_st_name, unknown_group_name, min_gene_count,
                                  header_function, args):
    seqs = data_folder + '/' + alleles_fasta
    database = data_folder + '/' + profiles_txt
    mlst_header = header_function()
    max_missing = len(mlst_header) - min_gene_count

    st, st_detail, group, spurious_hits = \
        mlst_blast(seqs, database, 'yes', [contigs], min_cov=args.min_coverage,
                   min_ident=args.min_identity, max_missing=max_missing, check_for_truncation=True,
                   report_incomplete=True, allow_multiple=True,
                   min_gene_count=min_gene_count, unknown_group_name=unknown_group_name,
                   min_spurious_cov=args.min_spurious_coverage,
                   min_spurious_ident=args.min_spurious_identity)

    assert len(mlst_header) == len(st_detail)

    results = {vir_name: group,
               vir_st_name: st}
    results.update(dict(zip(mlst_header, st_detail)))
    results['spurious_virulence_hits'] = ';'.join(spurious_hits)
    return results


def get_all_virulence_results(data_folder, contigs, args):
    virulence_results = [get_ybt_mlst_results(data_folder, contigs, args),
                         get_clb_mlst_results(data_folder, contigs, args),
                         get_iuc_mlst_results(data_folder, contigs, args),
                         get_iro_mlst_results(data_folder, contigs, args),
                         get_rmp_mlst_results(data_folder, contigs, args)]

    # Merge the results from different loci together. All columns should be unique between them
    # except for the 'spurious_virulence_hits' column which will be common to all of them.
    results = {'spurious_virulence_hits': []}
    for r in virulence_results:
        for column, val in r.items():
            if column not in results:
                results[column] = val
            elif column == 'spurious_virulence_hits':
                if val != '':
                    results['spurious_virulence_hits'].append(val)
            else:
                assert False

    if len(results['spurious_virulence_hits']) == 0:
        results['spurious_virulence_hits'] = '-'
    else:
        results['spurious_virulence_hits'] = ';'.join(results['spurious_virulence_hits'])
    return results


def get_ybt_mlst_results(data_folder, contigs, args):
    return get_virulence_cluster_results(data_folder, contigs, 'ybt_alleles.fasta',
                                         'YbST_profiles.txt', 'Yersiniabactin', 'YbST',
                                         'ybt unknown', 6, get_ybt_mlst_header, args)


def get_clb_mlst_results(data_folder, contigs, args):
    return get_virulence_cluster_results(data_folder, contigs, 'clb_alleles.fasta',
                                         'CbST_profiles.txt', 'Colibactin', 'CbST',
                                         'clb unknown', 8, get_clb_mlst_header, args)


def get_iuc_mlst_results(data_folder, contigs, args):
    return get_virulence_cluster_results(data_folder, contigs, 'iuc_alleles.fasta',
                                         'AbST_profiles.txt', 'Aerobactin', 'AbST',
                                         'iuc unknown', 3, get_iuc_mlst_header, args)


def get_iro_mlst_results(data_folder, contigs, args):
    return get_virulence_cluster_results(data_folder, contigs, 'iro_alleles.fasta',
                                         'SmST_profiles.txt', 'Salmochelin', 'SmST',
                                         'iro unknown', 2, get_iro_mlst_header, args)


def get_rmp_mlst_results(data_folder, contigs, args):
    return get_virulence_cluster_results(data_folder, contigs, 'rmp_alleles.fasta',
                                         'RmST_profiles.txt', 'RmpADC', 'RmST',
                                         'rmp unknown', 2, get_rmp_mlst_header, args)


def get_rmpa2_results(data_folder, contigs, args):
    seqs = data_folder + '/rmpA2.fasta'
    rmpa2_allele = rmpa2_blast(seqs, [contigs], args.min_coverage, args.min_identity)
    return {'rmpA2': rmpa2_allele}


def get_wzi_and_k_locus_results(data_folder, contigs, args):
    seqs = data_folder + '/wzi.fasta'
    database = data_folder + '/wzi.txt'
    bst, _, k_type, _ = mlst_blast(seqs, database, 'yes', [contigs], min_cov=args.min_coverage,
                                   min_ident=args.min_identity, max_missing=0, allow_multiple=False)
    if bst == '0':
        wzi_st = '-'
    else:
        wzi_st = 'wzi' + bst
    if k_type == '':
        k_type = '-'
    return {'wzi': wzi_st,
            'K_locus': k_type}


def get_resistance_results(data_folder, contigs, args, res_headers, kp_complex):
    if not pathlib.Path(contigs).is_file():
        raise OSError
    if args.resistance:
        gene_info, _, _ = read_class_file(data_folder + '/CARD_AMR_clustered.csv')

        # Only do mutation/truncation tests for Kp complex species.
        if kp_complex:
            qrdr = data_folder + '/QRDR_120.fasta'
            trunc = data_folder + '/MgrB_and_PmrB.fasta'
            omp = data_folder + '/OmpK.fasta'
        else:
            qrdr, trunc, omp = None, None, None

        seqs = data_folder + '/CARD_v3.1.13.fasta'
        res_hits = resblast_one_assembly(contigs, gene_info, qrdr, trunc, omp, seqs,
                                         args.min_coverage,  args.min_identity,
                                         args.min_spurious_coverage, args.min_spurious_identity)

        # Double check that there weren't any results without a corresponding output header.
        for h in res_hits.keys():
            if h not in res_headers:
                sys.exit( f'Error: results contained a value ({h}) that is not covered by the '
                          f'output headers')

        return {r: ';'.join(sorted(res_hits[r])) if r in res_hits else '-' for r in res_headers}
    else:
        return {}


def get_summary_results(results, res_headers, args):
    summary = {'virulence_score': str(get_virulence_score(results['Yersiniabactin'],
                                                          results['Colibactin'],
                                                          results['Aerobactin']))}
    if args.resistance:
        res_hits = [results[x] for x in res_headers]
        summary['resistance_score'] = str(get_resistance_score(res_headers, res_hits))
        summary['num_resistance_classes'] = str(get_resistance_class_count(res_headers, res_hits))
        summary['num_resistance_genes'] = str(get_resistance_gene_count(res_headers, res_hits))
    return summary


def output_headers(stdout_header, full_header, outfile):
    print('\t'.join(stdout_header))
    with open(outfile, 'wt') as o:
        o.write('\t'.join(full_header))
        o.write('\n')


def output_results(stdout_header, full_header, outfile, results):
    print('\t'.join([results[x] for x in stdout_header]))
    with open(outfile, 'at') as o:
        o.write('\t'.join([results[x] for x in full_header]))
        o.write('\n')

    # Double check that there weren't any results without a corresponding output header.
    for h in results.keys():
        if h not in full_header:
            sys.exit(f'Error: results contained a value ({h}) that is not covered by the output '
                     f'headers')


def get_strain_name(full_path):
    filename = os.path.split(full_path)[1]
    if filename.endswith('_temp_decompress.fasta'):
        filename = filename[:-22]
    if filename.endswith('.gz'):
        filename = filename[:-3]
    return os.path.splitext(filename)[0]


def rebuild_blast_indices(data_dir):
    for fasta in pathlib.Path(data_dir).glob('*.fasta'):
        makeblastdb_cmd = ['makeblastdb', '-dbtype', 'nucl', '-in', str(fasta)]
        with open(os.devnull, 'w') as devnull:
            subprocess.check_call(makeblastdb_cmd, stdout=devnull)


if __name__ == '__main__':
    main()



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import os
import subprocess

from .misc import reverse_complement


def run_blastn(db, query, min_cov, min_ident):
    build_blast_database_if_needed(db)

    cmd = 'blastn -task blastn -db {} -query {}'.format(db, query)
    cmd += " -outfmt '6 sacc pident slen length bitscore qseq sstrand sstart send" \
           " qacc qstart qend qframe'"
    cmd += ' -dust no -evalue 1E-20 -word_size 32 -max_target_seqs 10000'
    cmd += ' -perc_identity {}'.format(min_ident)

    # TODO: switch this over to subprocess
    blast_hits = []
    f = os.popen(cmd)
    for line in f:
        blast_hits.append(BlastHit(line))
    f.close()

    # Toss out low identity and low coverage hits.
    if min_ident is not None:
        blast_hits = [h for h in blast_hits if h.pcid * 100 >= min_ident]
    if min_cov is not None:
        blast_hits = [h for h in blast_hits if h.ref_cov * 100 >= min_cov]

    # Clean up redundant hits so we only have one hit for each part of the genome.
    blast_hits = cull_redundant_hits(blast_hits)

    # Get rid of any hits that start with 'delete_'. This is used in cases where there can be
    # confounding homologous hits. E.g. searching for rmpA2 alleles but not wanting to be misled by
    # homologous rmpA alleles.
    blast_hits = [h for h in blast_hits if not h.gene_id.startswith('delete_')]

    return blast_hits


def cull_redundant_hits(blast_hits):
    """
    Cull out redundant hits here (essentially implementing BLAST's -culling_limit 1 feature but
    with our own logic).
    """
    # Sort the hits from best to worst. Hit quality is defined as the product of gene coverage,
    # identity and score.
    blast_hits = sorted(blast_hits, key=lambda x: (1/(x.pcid * x.score * x.ref_cov), x.gene_id))

    filtered_blast_hits = []

    for h in blast_hits:
        if not overlapping(h, filtered_blast_hits):
            filtered_blast_hits.append(h)

    return filtered_blast_hits


def overlapping(hit, existing_hits):
    # Only consider hits in the same reading frame.
    existing_hits = [h for h in existing_hits if
                     h.strand == hit.strand and h.frame == hit.frame and
                     h.contig_name == hit.contig_name]

    for existing_hit in existing_hits:
        if hits_overlap(hit, existing_hit):
            return True

    return False


def hits_overlap(a, b):
    if a.contig_start <= b.contig_end and b.contig_start <= a.contig_end:  # There is some overlap
        allowed_overlap = 50
        overlap_size = len(range(max(a.contig_start, b.contig_start),
                                 min(a.contig_end, b.contig_end) + 1))
        return overlap_size > allowed_overlap
    else:
        return False


class BlastHit(object):
    def __init__(self, line):
        fields = line.rstrip().split('\t')
        self.gene_id = fields[0]                # sacc
        self.pcid = float(fields[1])            # pident
        self.ref_length = int(fields[2])        # slen
        self.alignment_length = int(fields[3])  # length
        self.score = float(fields[4])           # score
        self.hit_seq = fields[5]                # qseq
        self.strand = fields[6]                 # sstrand
        self.ref_start = int(fields[7])         # sstart
        self.ref_end = int(fields[8])           # send
        self.contig_name = fields[9]            # qacc
        self.contig_start = int(fields[10])     # qstart
        self.contig_end = int(fields[11])       # qend
        self.frame = fields[12]                 # qframe

        if self.strand == 'plus':
            assert self.ref_end >= self.ref_start
            self.ref_hit_len = self.ref_end - self.ref_start + 1
        else:
            assert self.strand == 'minus'
            assert self.ref_start >= self.ref_end
            self.ref_hit_len = self.ref_start - self.ref_end + 1
        assert self.contig_end >= self.contig_start

        self.ref_cov = self.ref_hit_len / self.ref_length

    def get_seq_start_end_pos_strand(self):
        # BLAST gives the aligned sequence, so we might need to remove dashes if there are
        # deletions relative to the reference.
        nucl_seq = self.hit_seq.replace('-', '')

        # BLAST also returns the contig's sequence so we might need to flip to the ref strand.
        if self.strand == 'minus':
            return reverse_complement(nucl_seq), self.ref_end, self.ref_start
        else:
            return nucl_seq, self.ref_start, self.ref_end


def build_blast_database_if_needed(seqs):
    if not os.path.exists(seqs + '.nin'):
        with open(os.devnull, 'w') as devnull:
            subprocess.check_call('makeblastdb -dbtype nucl -in ' + seqs, stdout=devnull,
                                  shell=True)



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from kleborate.__main__ import get_data_path
from kleborate.species import get_klebsiella_species, is_kp_complex


class TestSpecies(unittest.TestCase):

    def setUp(self):
        self.data_dir = get_data_path()

    def test_klebsiella_aerogenes(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000215745.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella aerogenes')

    def test_klebsiella_grimontii(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000733495.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella grimontii')

    def test_klebsiella_indica(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_005860775.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella indica')

    def test_klebsiella_michiganensis(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000240325.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella michiganensis')

    def test_klebsiella_oxytoca(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000247855.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella oxytoca')

    def test_klebsiella_pasteurii(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCA_902158585.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella pasteurii')

    def test_klebsiella_pneumoniae(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000016305.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella pneumoniae')

    def test_klebsiella_quasipneumoniae_subsp_quasipneumoniae(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000492415.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella quasipneumoniae subsp. quasipneumoniae')

    def test_klebsiella_quasipneumoniae_subsp_similipneumoniae(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000492795.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella quasipneumoniae subsp. similipneumoniae')

    def test_klebsiella_quasivariicola(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000523395.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella quasivariicola')

    def test_klebsiella_spallanzanii(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCA_901563875.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella spallanzanii')

    def test_klebsiella_variicola_subsp_variicola(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000019565.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella variicola subsp. variicola')

    def test_klebsiella_variicola_subsp_tropicalensis(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_002806645.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella variicola subsp. tropica')

    def test_klebsiella_africana(self):
        species, _ = get_klebsiella_species('test/test_genomes/ERR2835900.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella africana')

    def test_raoultella_planticola(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000648315.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella planticola')

    def test_raoultella_ornithinolytica(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000247895.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella ornithinolytica')

    def test_raoultella_terrigena(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_000829965.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Klebsiella terrigena')

    def test_salmonella(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_004010735.1.fna.gz',
                                            self.data_dir)
        self.assertTrue('Salmonella' in species)

    def test_citrobacter(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_003937345.1.fna.gz',
                                            self.data_dir)
        self.assertTrue('Citrobacter' in species)

    def test_yersinia_unknown(self):
        species, _ = get_klebsiella_species('test/test_genomes/GCF_001123825.1.fna.gz',
                                            self.data_dir)
        self.assertEqual(species, 'Yersinia (unknown species)')


class TestKpComplex(unittest.TestCase):

    def test_klebsiella_aerogenes(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella aerogenes'}))

    def test_klebsiella_grimontii(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella grimontii'}))

    def test_klebsiella_michiganensis(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella michiganensis'}))

    def test_klebsiella_oxytoca(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella oxytoca'}))

    def test_klebsiella_pneumoniae(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella pneumoniae'}))

    def test_klebsiella_quasipneumoniae_subsp_quasipneumoniae(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella quasipneumoniae subsp. '
                                                  'quasipneumoniae'}))

    def test_klebsiella_quasipneumoniae_subsp_similipneumoniae(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella quasipneumoniae subsp. '
                                                  'similipneumoniae'}))

    def test_klebsiella_quasivariicola(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella quasivariicola'}))

    def test_klebsiella_variicola_subsp_variicola(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella variicola subsp. variicola'}))

    def test_klebsiella_variicola_subsp_tropicalensis(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella variicola subsp. tropica'}))

    def test_klebsiella_africanensis(self):
        self.assertTrue(is_kp_complex({'species': 'Klebsiella africana'}))

    def test_raoultella_planticola(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella planticola'}))

    def test_raoultella_ornithinolytica(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella ornithinolytica'}))

    def test_raoultella_terrigena(self):
        self.assertFalse(is_kp_complex({'species': 'Klebsiella terrigena'}))

    def test_salmonella(self):
        self.assertFalse(is_kp_complex({'species': 'Salmonella enterica'}))

    def test_citrobacter(self):
        self.assertFalse(is_kp_complex({'species': 'Citrobacter freundii'}))



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import os
import subprocess
import sys
import tempfile


def get_kaptive_paths():
    this_file = os.path.realpath(__file__)
    kaptive_dir = os.path.join(os.path.dirname(os.path.dirname(this_file)), 'kaptive')
    if not os.path.isdir(kaptive_dir):
        sys.exit('Error: could not find Kaptive directory. Did you git clone with --recursive?')
    kaptive_py = os.path.join(kaptive_dir, 'kaptive.py')
    if not os.path.isfile(kaptive_py):
        sys.exit('Error: could not find kaptive.py')
    db_dir = os.path.join(kaptive_dir, 'reference_database')
    kaptive_k_db = os.path.join(db_dir, 'Klebsiella_k_locus_primary_reference.gbk')
    if not os.path.isfile(kaptive_k_db):
        sys.exit('Error: could not find Klebsiella_k_locus_primary_reference.gbk')
    kaptive_o_db = os.path.join(db_dir, 'Klebsiella_o_locus_primary_reference.gbk')
    if not os.path.isfile(kaptive_o_db):
        sys.exit('Error: could not find Klebsiella_o_locus_primary_reference.gbk')
    return kaptive_py, kaptive_k_db, kaptive_o_db


def get_kaptive_results(locus_type, kaptive_py, kaptive_db, contigs, args):
    assert locus_type == 'K' or locus_type == 'O'

    headers = ['K_locus', 'K_type', 'K_locus_confidence', 'K_locus_problems', 'K_locus_identity',
               'K_locus_missing_genes']
    if locus_type == 'O':
        headers = [x.replace('K_', 'O_') for x in headers]

    if (args.kaptive_k and locus_type == 'K') or (args.kaptive_o and locus_type == 'O'):
        if locus_type == 'K':
            outfile = args.kaptive_k_outfile
        else:  # locus_type == 'O':
            outfile = args.kaptive_o_outfile

        # Try with multiple threads, and if that fails (possible due to a multithreading BLAST
        # error), try again with one thread.
        kaptive_results = run_kaptive(kaptive_py, kaptive_db, contigs, outfile,
                                      args.min_kaptive_confidence, one_thread=False)
        if kaptive_results is None:
            kaptive_results = run_kaptive(kaptive_py, kaptive_db, contigs, outfile,
                                          args.min_kaptive_confidence, one_thread=True)

        assert len(headers) == len(kaptive_results)
        return dict(zip(headers, kaptive_results))
    else:
        return {}


def run_kaptive(kaptive_py, kaptive_db, contigs, output_file, min_confidence, one_thread):
    thread_option = ' --threads 1' if one_thread else ''
    with tempfile.TemporaryDirectory() as tmp_dir:
        kaptive_prefix = tmp_dir + '/kaptive'
        kaptive_table = kaptive_prefix + '_table.txt'

        p = subprocess.Popen(sys.executable + ' ' + kaptive_py +
                             ' -a ' + contigs +
                             ' -k ' + kaptive_db +
                             ' -o ' + kaptive_prefix +
                             ' --verbose --no_seq_out --no_json' +
                             thread_option,
                             shell=True,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()

        # Make sure the output is a string, whether we are in Python 2 or 3.
        if not isinstance(stdout, str):
            stdout = stdout.decode()
        if not isinstance(stderr, str):
            stderr = stderr.decode()

        # If we hit the BLAST threading problem, return None and we'll try again with one thread.
        if 'tblastn crashed!' in stderr and not one_thread:
            return None

        if p.returncode != 0:
            if stderr:
                sys.exit('Error: Kaptive failed to run with the following error:\n' +
                         stderr.strip())
            else:
                sys.exit('Error: Kaptive failed to run')

        locus, serotype, confidence, problems, identity = None, None, None, None, None
        missing = []

        # Parse the required information from the Kaptive verbose output.
        output_lines = stdout.splitlines()
        missing_gene_lines = False
        for line in output_lines:
            if 'Best match locus:' in line:
                locus = line.split('Best match locus:')[1].strip()
            if 'Best match type:' in line:
                serotype = line.split('Best match type:')[1].strip()
            if 'Match confidence:' in line:
                confidence = line.split('Match confidence:')[1].strip()
            if 'Problems:' in line:
                problems = line.split('Problems:')[1].strip()
                if problems == 'None':
                    problems = problems.lower()
            if 'Identity:' in line:
                identity = line.split('Identity:')[1].strip()
            if 'Other genes in locus:' in line:
                missing_gene_lines = False
            if missing_gene_lines:
                missing_gene = line.strip()
                if missing_gene:
                    missing.append(missing_gene)
            if 'Missing expected genes:' in line:
                missing_gene_lines = True

        if output_file:	 # if we are saving Kaptive results to file...
            with open(kaptive_table, 'rt') as f:
                kaptive_table_lines = f.readlines()
            assert len(kaptive_table_lines) == 2
            if not os.path.isfile(output_file):
                with open(output_file, 'wt') as f:
                    f.write(kaptive_table_lines[0])	 # write header line
            with open(output_file, 'at') as f:
                f.write(kaptive_table_lines[1])		 # write data line

    if locus is None or serotype is None or confidence is None or problems is None or identity is None:
        sys.exit('Error: Kaptive failed to produce the expected output')

    if not confidence_meets_threshold(confidence, min_confidence):
        locus = f'unknown ({locus})'
        if not serotype.startswith('unknown'):
            serotype = f'unknown ({serotype})'

    return [locus, serotype, confidence, problems, identity, ','.join(missing)]


def confidence_meets_threshold(confidence, min_confidence):
    """
    Returns True if the confidence level meets or exceeds the minimum confidence level.
    """
    min_confidence = min_confidence.replace('_', ' ')
    scores = {'None': 0, 'Low': 1, 'Good': 2, 'High': 3, 'Very high': 4, 'Perfect': 5}
    return scores[confidence] >= scores[min_confidence]



"""
Reports best match
  * :    best matching allele is not precise match
  -nLV : best matching ST is n-locus variant

If an annotation column is provided (such as clonal complex) in the final column of the profiles
file, this annotation will be reported in column 2 of the output table

NOTE there is a bug with the culling_limit parameter in older versions of BLAST+. This code has
been tested with BLAST+2.2.30. It does not work with BLAST2.2.25. Not sure about other versions.

Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

from .blastn import run_blastn
from .truncation import truncation_check


def rmpa2_blast(seqs, assemblies, min_cov, min_ident):
    for contigs in assemblies:
        rmpa2_calls = []
        hits = run_blastn(seqs, contigs, min_cov, min_ident)
        for hit in hits:
            if hit.alignment_length > hit.ref_length / 2:
                gene_id = hit.gene_id
                if hit.pcid < 100.00 or hit.alignment_length < hit.ref_length:
                    gene_id += '*'
                gene_id += truncation_check(hit)[0]
                rmpa2_calls.append(gene_id)
        if len(rmpa2_calls) == 0:
            rmpa2_calls.append('-')
        return ','.join(rmpa2_calls)



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_output_headers, get_resistance_results
from kleborate.shv_mutations import get_new_bla_class, get_class_changing_mutations


class TestShvMutations(unittest.TestCase):
    """
    Tests SHV mutation logic.
    """

    def setUp(self):
        self.data_dir = 'test/test_shv/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def test_shv_01(self):
        """
        This test has an exact match for SHV-1.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/01.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-1')
        self.assertEqual(results['SHV_mutations'], '-')

    def test_shv_02(self):
        """
        This test has a match for SHV-1 with a mutation at site 238 (G -> Y). This changes the
        class to ESBL.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/02.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_ESBL_acquired'], 'SHV-1* +238Y')
        self.assertEqual(results['SHV_mutations'], '238Y')

    def test_shv_03(self):
        """
        Same as test 2, but the gene is on the reverse strand.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/03.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_ESBL_acquired'], 'SHV-1* +238Y')
        self.assertEqual(results['SHV_mutations'], '238Y')

    def test_shv_04(self):
        """
        This test has a match for SHV-1 with a mutation at site 50 (G -> Y). This is not class-modifying 
        and so won't be reported.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/04.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-1*')
        self.assertEqual(results['SHV_mutations'], '-')

    def test_shv_05(self):
        """
        This test has an exact match for SHV-29.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/05.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_ESBL_acquired'], 'SHV-29')
        self.assertEqual(results['SHV_mutations'], '238A;35Q')

    def test_shv_06(self):
        """
        This test has SHV-29 plus an beta-lactamase inhibitor resistant mutation.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/06.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_ESBL_inhR_acquired'], 'SHV-29* +234Y')
        self.assertEqual(results['SHV_mutations'], '234Y;238A;35Q')

    def test_shv_07(self):
        """
        This test has SHV-1 with position 238 deleted. Is not 
        reported and doesn't have an effect.
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/07.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-1*')
        self.assertEqual(results['SHV_mutations'], '-')

    def test_shv_08(self):
        """
        This test has SHV-1 with a synonymous mutation in the omega loop (not ESBL-conferring mutation).
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/08.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-1^')
        self.assertEqual(results['SHV_mutations'], '-')

    def test_shv_09(self):
        """
        This test has SHV-1 with a nonsynonymous mutation in the omega loop (reported but not ESBL-conferring mutation).
        """
        results = get_resistance_results(self.data_dir, 'test/test_shv/09.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-1*')
        self.assertEqual(results['SHV_mutations'], '174R;omega-loop=RWETELNEALRGDARD')

    def test_bla_class_01(self):
        self.assertEqual(get_new_bla_class(False, False), 'Bla_chr')

    def test_bla_class_02(self):
        self.assertEqual(get_new_bla_class(True, False), 'Bla_ESBL')

    def test_bla_class_03(self):
        self.assertEqual(get_new_bla_class(False, True), 'Bla_inhR')

    def test_bla_class_04(self):
        self.assertEqual(get_new_bla_class(True, True), 'Bla_ESBL_inhR')

    def test_get_class_changing_mutations_01(self):
        mutations = get_class_changing_mutations('Bla_chr', 'Bla_chr', ['A'], ['B'])
        self.assertEqual(mutations, [])

    def test_get_class_changing_mutations_02(self):
        mutations = get_class_changing_mutations('Bla_ESBL', 'Bla_ESBL', ['A'], ['B'])
        self.assertEqual(mutations, [])

    def test_get_class_changing_mutations_03(self):
        mutations = get_class_changing_mutations('Bla_chr', 'Bla_ESBL', ['A'], ['B'])
        self.assertEqual(mutations, ['A'])

    def test_get_class_changing_mutations_04(self):
        mutations = get_class_changing_mutations('Bla_ESBL', 'Bla_chr', ['A'], ['B'])
        self.assertEqual(mutations, ['A'])

    def test_get_class_changing_mutations_05(self):
        mutations = get_class_changing_mutations('Bla_chr', 'Bla_inhR', ['A'], ['B'])
        self.assertEqual(mutations, ['B'])

    def test_get_class_changing_mutations_06(self):
        mutations = get_class_changing_mutations('Bla_inhR', 'Bla_chr', ['A'], ['B'])
        self.assertEqual(mutations, ['B'])

    def test_get_class_changing_mutations_07(self):
        mutations = get_class_changing_mutations('Bla_chr', 'Bla_ESBL_inhR', ['A'], ['B'])
        self.assertEqual(mutations, ['A', 'B'])

    def test_get_class_changing_mutations_08(self):
        mutations = get_class_changing_mutations('Bla_ESBL_inhR', 'Bla_chr', ['A'], ['B'])
        self.assertEqual(mutations, ['A', 'B'])



"""
Copyright 2023 Kat Holt (drkatholt@gmail.com)
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

from Bio.Seq import Seq
from Bio import pairwise2
from Bio.Align import substitution_matrices


def check_for_shv_mutations(hit, hit_allele, bla_class, exact_match):
    # Don't do anything on non-SHV genes.
    if 'SHV' not in hit_allele:
        return bla_class, [], [], None

    nucl_seq, _, _ = hit.get_seq_start_end_pos_strand()

    # If there are any ambiguous bases in the sequence, then we can't do the translation.
    ambiguous_bases = set(b for b in nucl_seq) - {'A', 'C', 'G', 'T'}
    if ambiguous_bases:
        return bla_class, [], [], None

    # BioPython doesn't like it if the sequence isn't a multiple of 3.
    nucl_seq = nucl_seq[:len(nucl_seq) // 3 * 3]

    coding_dna = Seq(nucl_seq)
    translation = str(coding_dna.translate(table='Bacterial', to_stop=True))

    shv_1_ref = 'MRYIRLCIISLLATLPLAVHASPQPLEQIKLSESQLSGRVGMIEMDLASGRTLTAWRADERFPMMSTFKVVLCGAVLAR' \
                'VDAGDEQLERKIHYRQQDLVDYSPVSEKHLADGMTVGELCAAAITMSDNSAANLLLATVGGPAGLTAFLRQIGDNVTRL' \
                'DRWETELNEALPGDARDTTTPASMAATLRKLLTSQRLSARSQRQLLQWMVDDRVAGPLIRSVLPAGWFIADKTGAGERG' \
                'ARGIVALLGPNNKAERIVVIYLRDTPASMAERNQQIAGIGAALIEHWQR'

    blosum62 = substitution_matrices.load('BLOSUM62')
    alignments = pairwise2.align.globalds(shv_1_ref, translation, blosum62, -10, -0.5)

    # If we didn't get any global amino acid alignments, then it's not appropriate to look for SHV
    # mutations in this hit.
    if not alignments:
        return bla_class, [], [], None

    alignment = alignments[0]
    ref_aligned, hit_aligned, score, _, _ = alignment

    # If the identity of the alignment is too low, then it's not appropriate to look for SHV
    # mutations in this hit.
    identity = get_percent_identity(ref_aligned, hit_aligned)
    if identity < 0.9:
        return bla_class, [], [], None

    
    
    # Mutations at these sites will lead to an ESBL class:
    pos_169_mut, pos_169_aa = get_mut(ref_aligned, hit_aligned, 164, 169, 'L')
    pos_179_mut, pos_179_aa = get_mut(ref_aligned, hit_aligned, 174, 179, 'D')
    pos_238_mut, pos_238_aa = get_mut(ref_aligned, hit_aligned, 233, 238, 'G')
    pos_148_mut, pos_148_aa = get_mut(ref_aligned, hit_aligned, 143, 148, 'L')
    
    # Mutations at site Ambler-240 will lead to an ESBL class, but only if site Ambler-35 is also mutated.
    pos_035_mut, pos_035_aa = get_mut(ref_aligned, hit_aligned,  30,  35, 'L')
    pos_240_mut, pos_240_aa = get_mut(ref_aligned, hit_aligned, 234, 240, 'E')

    has_esbl = (pos_169_mut or pos_179_mut or pos_238_mut or pos_148_mut or 
    										(pos_240_mut and pos_035_mut))

    esbl_mutations = [pos_169_mut, pos_179_mut, pos_238_mut, pos_148_mut, pos_240_mut]

    # Mutations at these sites will lead to inhibition:
    pos_069_mut, pos_069_aa = get_mut(ref_aligned, hit_aligned,  64,  69, 'M')
    pos_130_mut, pos_130_aa = get_mut(ref_aligned, hit_aligned, 125, 130, 'S')
    pos_234_mut, pos_234_aa = get_mut(ref_aligned, hit_aligned, 229, 234, 'K')
    pos_235_mut, pos_235_aa = get_mut(ref_aligned, hit_aligned, 230, 235, 'T')

    has_inhr = (pos_069_mut or pos_130_mut or pos_234_mut or pos_235_mut)
    inhr_mutations = [pos_069_mut, pos_130_mut, pos_234_mut, pos_235_mut]

    # Mutations at these sites don't change the class, but will still be reported:
    pos_025_mut, pos_025_aa = get_mut(ref_aligned, hit_aligned,  20,  25, 'A')
    pos_035_mut, pos_035_aa = get_mut(ref_aligned, hit_aligned,  30,  35, 'L')
    pos_146_mut, pos_146_aa = get_mut(ref_aligned, hit_aligned, 141, 146, 'A')
    pos_156_mut, pos_156_aa = get_mut(ref_aligned, hit_aligned, 151, 156, 'G')

    # Mutations in the omega loop, for tracking only (not class modification):
    pos_164_mut, pos_164_aa = get_mut(ref_aligned, hit_aligned, 159, 164, 'R')
    pos_165_mut, pos_165_aa = get_mut(ref_aligned, hit_aligned, 160, 165, 'W')
    pos_166_mut, pos_166_aa = get_mut(ref_aligned, hit_aligned, 161, 166, 'E')
    pos_167_mut, pos_167_aa = get_mut(ref_aligned, hit_aligned, 162, 167, 'T')
    pos_168_mut, pos_168_aa = get_mut(ref_aligned, hit_aligned, 163, 168, 'E')
    pos_170_mut, pos_170_aa = get_mut(ref_aligned, hit_aligned, 165, 170, 'N')
    pos_171_mut, pos_171_aa = get_mut(ref_aligned, hit_aligned, 166, 171, 'E')
    pos_172_mut, pos_172_aa = get_mut(ref_aligned, hit_aligned, 167, 172, 'A')
    pos_173_mut, pos_173_aa = get_mut(ref_aligned, hit_aligned, 168, 173, 'L')
    pos_174_mut, pos_174_aa = get_mut(ref_aligned, hit_aligned, 169, 174, 'P')
    pos_175_mut, pos_175_aa = get_mut(ref_aligned, hit_aligned, 170, 175, 'G')
    pos_176_mut, pos_176_aa = get_mut(ref_aligned, hit_aligned, 171, 176, 'D')
    pos_177_mut, pos_177_aa = get_mut(ref_aligned, hit_aligned, 172, 177, 'A')
    pos_178_mut, pos_178_aa = get_mut(ref_aligned, hit_aligned, 173, 178, 'R')

    omega_loop_seq = ''.join([pos_164_aa, pos_165_aa, pos_166_aa, pos_167_aa, pos_168_aa,
                              pos_169_aa, pos_170_aa, pos_171_aa, pos_172_aa, pos_173_aa,
                              pos_174_aa, pos_175_aa, pos_176_aa, pos_177_aa, pos_178_aa,
                              pos_179_aa])
                              
    if omega_loop_seq == 'RWETELNEALPGDARD':  # if it's the same as SHV-1
        omega_loop_seq = None

    shv_mutations = [pos_025_mut, pos_035_mut, pos_069_mut, pos_130_mut, pos_146_mut, pos_148_mut,
                     pos_156_mut, pos_164_mut, pos_165_mut, pos_166_mut, pos_167_mut, pos_168_mut,
                     pos_169_mut, pos_170_mut, pos_171_mut, pos_172_mut, pos_173_mut, pos_174_mut,
                     pos_175_mut, pos_176_mut, pos_177_mut, pos_178_mut, pos_179_mut, pos_234_mut,
                     pos_235_mut, pos_238_mut, pos_240_mut]
                     
    shv_mutations = [m for m in shv_mutations if m]

    if exact_match:
        new_bla_class = bla_class
    else:
        new_bla_class = get_new_bla_class(has_esbl, has_inhr)
    class_changing_mutations = \
        get_class_changing_mutations(bla_class, new_bla_class, esbl_mutations, inhr_mutations)

    return new_bla_class, shv_mutations, class_changing_mutations, omega_loop_seq


def get_percent_identity(ref_aligned, hit_aligned):
    matches = 0
    for i, a in enumerate(ref_aligned):
        b = hit_aligned[i]
        if a == b:
            matches += 1
    return matches / len(ref_aligned)


def get_mut(ref_aligned, hit_aligned, ref_pos, ambler_pos, ref_aa):
    """
    ref_pos:    the index of the AA in SHV-1 (0-based because we're looking it up in Python)
    ambler_pos: the index of the AA in the Ambler alignment (1-based because it's just for the name)
    ref_aa:     the AA in the SHV-1 sequence
    """
    ref_no_gaps, hit_no_gaps = [], []
    for i, a in enumerate(ref_aligned):
        b = hit_aligned[i]
        if a != '-':
            ref_no_gaps.append(a)
            hit_no_gaps.append(b)
    assert len(ref_no_gaps) == 286
    assert ref_no_gaps[ref_pos] == ref_aa
    hit_aa = hit_no_gaps[ref_pos]
    if ref_aa != hit_aa and hit_aa != '-':
        mutation_notation = f'{ambler_pos}{hit_aa}'
    else:
        mutation_notation = ''
    return mutation_notation, hit_aa


def get_new_bla_class(esbl, inhr):
    if not esbl and not inhr:
        return 'Bla_chr'
    if esbl and not inhr:
        return 'Bla_ESBL'
    if inhr and not esbl:
        return 'Bla_inhR'
    if esbl and inhr:
        return 'Bla_ESBL_inhR'


def get_class_changing_mutations(bla_class, new_bla_class, esbl_mutations, inhr_mutations):
    if bla_class == new_bla_class:
        return []
    class_changing_mutations = []
    if ('ESBL' in bla_class and 'ESBL' not in new_bla_class) or \
            ('ESBL' not in bla_class and 'ESBL' in new_bla_class):
        class_changing_mutations += esbl_mutations
    if ('inhR' in bla_class and 'inhR' not in new_bla_class) or \
            ('inhR' not in bla_class and 'inhR' in new_bla_class):
        class_changing_mutations += inhr_mutations
    return [m for m in class_changing_mutations if m]



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import gzip
import sys


def load_fasta(filename):
    """
    Returns the names and sequences for the given fasta file as a list of tuples (name, seq).
    """
    fasta_seqs = []
    with get_open_func(filename)(filename, 'rt') as fasta_file:
        name = ''
        sequence = ''
        for line in fasta_file:
            line = line.strip()
            if not line:
                continue
            if line[0] == '>':  # Header line = start of new contig
                if name:
                    fasta_seqs.append((name.split()[0], sequence.upper()))
                    sequence = ''
                name = line[1:]
            else:
                sequence += line
        if name:
            fasta_seqs.append((name.split()[0], sequence.upper()))
    return fasta_seqs


def get_compression_type(filename):
    """
    Attempts to guess the compression (if any) on a file using the first few bytes.
    http://stackoverflow.com/questions/13044562
    """
    magic_dict = {'gz': (b'\x1f', b'\x8b', b'\x08'),
                  'bz2': (b'\x42', b'\x5a', b'\x68'),
                  'zip': (b'\x50', b'\x4b', b'\x03', b'\x04')}
    max_len = max(len(x) for x in magic_dict)
    unknown_file = open(filename, 'rb')
    file_start = unknown_file.read(max_len)
    unknown_file.close()
    compression_type = 'plain'
    for file_type, magic_bytes in magic_dict.items():
        if file_start.startswith(magic_bytes):
            compression_type = file_type
    if compression_type == 'bz2':
        sys.exit('Error: cannot use bzip2 format - use gzip instead')
    if compression_type == 'zip':
        sys.exit('Error: cannot use zip format - use gzip instead')
    return compression_type


def get_open_func(filename):
    if get_compression_type(filename) == 'gz':
        return gzip.open
    else:  # plain text
        return open


REV_COMP_DICT = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'a': 't', 't': 'a', 'g': 'c', 'c': 'g',
                 'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W', 'K': 'M', 'M': 'K', 'B': 'V', 'V': 'B',
                 'D': 'H', 'H': 'D', 'N': 'N', 'r': 'y', 'y': 'r', 's': 's', 'w': 'w', 'k': 'm',
                 'm': 'k', 'b': 'v', 'v': 'b', 'd': 'h', 'h': 'd', 'n': 'n', '.': '.', '-': '-',
                 '?': '?'}


def complement_base(base):
    try:
        return REV_COMP_DICT[base]
    except KeyError:
        return 'N'


def reverse_complement(seq):
    return ''.join([complement_base(x) for x in seq][::-1])



"""
The version is stored here in a separate file so it can exist in only one place.
http://stackoverflow.com/questions/458550

Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""


__version__ = '2.4.0'



#!/usr/bin/env python3
"""
Convenience wrapper for running Kleborate directly from source tree.

Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

from kleborate.__main__ import main


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""
This script modifies the ARGannot_r3.fasta file (comes with SRST2) for use with Kleborate.
Specifically, it removes alleles which are not in the ARGannot_clustered80_r3.csv file.

It takes two arguments:
  1) SRST2's copy of ARGannot_r3.fasta
  2) the ARGannot_clustered80_r3.csv file made by make_argannot_csv.py

Usage:
  ./make_argannot_fasta.py path/to/srst2/data/ARGannot_r3.fasta ARGannot_clustered80_r3.csv > ARGannot_r3.fasta
"""

import sys


def main():
    argannot_fasta_filename = sys.argv[1]
    argannot_csv_filename = sys.argv[2]

    allele_names = set()
    with open(argannot_csv_filename, 'rt') as argannot_csv_file:
        for line in argannot_csv_file:
            allele_names.add(line.split(',')[4])

    fasta_seqs = load_fasta(argannot_fasta_filename)
    fasta_seqs = sorted(fasta_seqs, key=fasta_sorting_key)

    for name, seq in fasta_seqs:
        name = name.replace('NimB_Nitroimidazole_Gene', 'NimB')  # fix verbose gene name
        name = name.replace('_Colistin__', '_Col__')  # colistin class name to just three letters
        name = name.replace('_Agly__', '_AGly__')  # fix capitalisation
        allele_name = name.split('__')[2]
        if allele_name in allele_names:
            print('>' + name)
            print(seq, end='')


def load_fasta(filename):
    """Returns the names and sequences for the given fasta file."""
    fasta_seqs = []
    with open(filename, 'rt') as fasta_file:
        name = ''
        sequence = ''
        for line in fasta_file:
            stripped_line = line.strip()
            if not stripped_line:
                continue
            if stripped_line[0] == '>':  # Header line = start of new contig
                if name:
                    fasta_seqs.append((name, sequence))
                    sequence = ''
                name = stripped_line[1:]
            else:
                sequence += line
        if name:
            fasta_seqs.append((name, sequence))
    return fasta_seqs


def fasta_sorting_key(fasta_record):
    """
    This function defines a sorting key for fasta so 'TEM-2' will come before 'TEM-10', etc.
    """
    name_parts = fasta_record[0].split()[0].split('__')
    return int(name_parts[0]), int(name_parts[3])


if __name__ == '__main__':
    main()



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest
from kleborate.__main__ import get_data_path, get_output_headers, get_summary_results


class TestResScore(unittest.TestCase):

    def setUp(self):
        self.data_dir = get_data_path()
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)
        self.results = {res_class: '-' for res_class in self.res_headers}
        self.results['Yersiniabactin'] = '-'
        self.results['Colibactin'] = '-'
        self.results['Aerobactin'] = '-'

    def test_nothing(self):
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '0')
        self.assertEqual(summary_results['resistance_score'], '0')
        self.assertEqual(summary_results['num_resistance_classes'], '0')
        self.assertEqual(summary_results['num_resistance_genes'], '0')

    def test_res_counts_1(self):
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_acquired'] = 'b'
        self.results['Tet_acquired'] = 'c'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '3')

    def test_res_counts_2(self):
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_acquired'] = 'b;c'
        self.results['Tet_acquired'] = 'd;e;f'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '6')

    def test_res_counts_3(self):
        """
        Intrinsic Bla genes should not add to the counts.
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_acquired'] = 'b;c'
        self.results['Tet_acquired'] = 'd;e;f'
        self.results['Bla_acquired'] = 'g'
        self.results['Bla_chr'] = 'h'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '4')
        self.assertEqual(summary_results['num_resistance_genes'], '7')

    def test_res_counts_4(self):
        """
        Bla genes in columns other than 'Bla_chr' (e.g. 'Bla_ESBL_acquired') should add to the
        counts.
        """
        self.results['Tet_acquired'] = 'a;b;c'
        self.results['Bla_chr'] = 'd'
        self.results['Bla_ESBL_acquired'] = 'e'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '2')
        self.assertEqual(summary_results['num_resistance_genes'], '4')

    def test_res_counts_5(self):
        """
        Bla genes in columns other than 'Bla_chr' (e.g. 'Bla_acquired') should add to the counts.
        """
        self.results['Tet_acquired'] = 'a;b;c'
        self.results['Bla_acquired'] = 'd'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '2')
        self.assertEqual(summary_results['num_resistance_genes'], '4')

    def test_res_counts_6(self):
        """
        Omp gene truncations should not add to the gene or class counts.
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_acquired'] = 'b;c'
        self.results['Tet_acquired'] = 'd;e;f'
        self.results['Omp_mutations'] = 'g;h'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '6')

    def test_res_counts_7(self):
        """
        Mutations should add to the class count but not to the gene count (because that is for
        acquired resistance genes).
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_mutations'] = 'b'
        self.results['Tet_acquired'] = 'c;d;e'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '4')

    def test_res_counts_8(self):
        """
        Mutations should add to the class count but not to the gene count (because that is for
        acquired resistance genes).
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_mutations'] = 'b;c'
        self.results['Tet_acquired'] = 'd;e;f'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '4')

    def test_res_counts_9(self):
        """
        Mutations should add to the class count but not to the gene count (because that is for
        acquired resistance genes).
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Tet_acquired'] = 'b;c'
        self.results['Col_mutations'] = 'd'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '3')

    def test_res_counts_10(self):
        """
        Resistance-by-truncation should add to the class count but not to the gene count (because
        that is for acquired resistance genes).
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Tet_acquired'] = 'b'
        self.results['Col_acquired'] = 'c'
        self.results['Col_mutations'] = 'd'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '3')

    def test_res_counts_11(self):
        """
        Spurious hits should not add to either the gene or class counts.
        """
        self.results['AGly_acquired'] = 'a'
        self.results['Flq_acquired'] = 'b'
        self.results['Tet_acquired'] = 'c'
        self.results['spurious_resistance_hits'] = 'd;e;f'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['num_resistance_classes'], '3')
        self.assertEqual(summary_results['num_resistance_genes'], '3')

    def test_res_score_1(self):
        self.results['Bla_chr'] = 'a'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '0')

    def test_res_score_2(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_ESBL_acquired'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '1')

    def test_res_score_3(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_ESBL_inhR_acquired'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '1')

    def test_res_score_4(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_Carb_acquired'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '2')

    def test_res_score_5(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_ESBL_acquired'] = 'b'
        self.results['Bla_Carb_acquired'] = 'c'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '2')

    def test_res_score_6(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_ESBL_acquired'] = 'b'
        self.results['Bla_Carb_acquired'] = 'c'
        self.results['Col_acquired'] = 'd'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '3')

    def test_res_score_7(self):
        self.results['Bla_chr'] = 'a'
        self.results['Bla_ESBL_acquired'] = 'b'
        self.results['Bla_Carb_acquired'] = 'c'
        self.results['Col_mutations'] = 'd'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '3')

    def test_res_score_8(self):
        self.results['Col_acquired'] = 'a'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['resistance_score'], '0')

    def test_vir_score_1(self):
        self.results['Yersiniabactin'] = 'a'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '1')

    def test_vir_score_2(self):
        self.results['Colibactin'] = 'a'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '2')

    def test_vir_score_3(self):
        self.results['Colibactin'] = 'a'
        self.results['Yersiniabactin'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '2')

    def test_vir_score_4(self):
        self.results['Aerobactin'] = 'a'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '3')

    def test_vir_score_5(self):
        self.results['Aerobactin'] = 'a'
        self.results['Yersiniabactin'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '4')

    def test_vir_score_6(self):
        self.results['Aerobactin'] = 'a'
        self.results['Colibactin'] = 'b'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '5')

    def test_vir_score_7(self):
        self.results['Aerobactin'] = 'a'
        self.results['Colibactin'] = 'b'
        self.results['Yersiniabactin'] = 'c'
        summary_results = get_summary_results(self.results, self.res_headers, self.args)
        self.assertEqual(summary_results['virulence_score'], '5')



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import unittest

from kleborate.kaptive import confidence_meets_threshold


class TestKaptive(unittest.TestCase):

    def test_confidence_meets_threshold_01(self):
        self.assertTrue(confidence_meets_threshold('None', 'None'))

    def test_confidence_meets_threshold_02(self):
        self.assertTrue(confidence_meets_threshold('Good', 'Good'))

    def test_confidence_meets_threshold_03(self):
        self.assertTrue(confidence_meets_threshold('Very high', 'Very_high'))

    def test_confidence_meets_threshold_04(self):
        self.assertTrue(confidence_meets_threshold('Perfect', 'Perfect'))

    def test_confidence_meets_threshold_05(self):
        self.assertTrue(confidence_meets_threshold('Perfect', 'Good'))

    def test_confidence_meets_threshold_06(self):
        self.assertTrue(confidence_meets_threshold('Perfect', 'Very_high'))

    def test_confidence_meets_threshold_07(self):
        self.assertFalse(confidence_meets_threshold('Very high', 'Perfect'))

    def test_confidence_meets_threshold_08(self):
        self.assertFalse(confidence_meets_threshold('High', 'Very_high'))

    def test_confidence_meets_threshold_09(self):
        self.assertFalse(confidence_meets_threshold('Good', 'High'))

    def test_confidence_meets_threshold_10(self):
        self.assertFalse(confidence_meets_threshold('Low', 'Good'))



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_output_headers, get_resistance_results


class TestResGyrAParC(unittest.TestCase):
    """
    Tests calling of fluoroquinolone resistance via snps in the GyrA and ParC genes.
    """

    def setUp(self):
        self.data_dir = 'test/test_res_qrdr/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def test_no_mutations(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_qrdr/test_res_qrdr_1.fasta',
                                         self.args, self.res_headers, True)
        self.assertEqual(results['Flq_mutations'], '-')

    def test_gyra(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_qrdr/test_res_qrdr_2.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('GyrA-83C' in results['Flq_mutations'])

    def test_parc(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_qrdr/test_res_qrdr_3.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('ParC-84D' in results['Flq_mutations'])

    def test_gyra_and_parc(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_qrdr/test_res_qrdr_4.fasta',
                                         self.args, self.res_headers, True)
        self.assertTrue('GyrA-83C' in results['Flq_mutations'])
        self.assertTrue('ParC-84D' in results['Flq_mutations'])

    def test_gyra_and_parc_non_kp_complex(self):
        """
        Setting the Kp complex variable to False should turn off the QRDR tests.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_qrdr/test_res_qrdr_4.fasta',
                                         self.args, self.res_headers, False)
        self.assertTrue('GyrA-83C' not in results['Flq_mutations'])
        self.assertTrue('ParC-84D' not in results['Flq_mutations'])



#!/usr/bin/env python3
"""
Reports best match
  * :   best matching allele is not precise match
 -nLV : best matching ST is n-locus variant

If an annotation column is provided (such as clonal complex) in the final column of the profiles
file, this annotation will be reported in column 2 of the output table.

Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import os
import re

from .blastn import run_blastn
from .truncation import truncation_check


def mlst_blast(seqs, database, info_arg, assemblies, min_cov, min_ident, max_missing,
               check_for_truncation=False, report_incomplete=False, allow_multiple=False,
               min_gene_count=None, unknown_group_name=None,
               min_spurious_cov=None, min_spurious_ident=None):
    st_names, alleles_to_st, st_to_info, header = load_st_database(database, info_arg)

    # In order to call an ST, there needs to be an exact match for half (rounded down) of the
    # alleles.
    required_exact_matches = int(len(header) / 2)

    contigs = assemblies[0]
    _, filename = os.path.split(contigs)
    name, _ = os.path.splitext(filename)

    if min_spurious_cov is not None:
        hits = run_blastn(seqs, contigs, min_spurious_cov, min_spurious_ident)
        num_hits_before = len(hits)
        spurious_hits = [h for h in hits
                         if h.ref_cov * 100 < min_cov or h.pcid * 100 < min_ident]
        hits = [h for h in hits if h.ref_cov * 100 >= min_cov and h.pcid * 100 >= min_ident]
        assert len(hits) + len(spurious_hits) == num_hits_before
    else:
        hits = run_blastn(seqs, contigs, min_cov, min_ident)
        spurious_hits = None

    final_call = ''
    final_alleles = [''] * len(header)
    final_info = ''

    # See if we have any loci for which there are multiple genes.
    hits_one_per_locus = keep_only_one_hit_per_locus(hits)
    any_multiple_hits = len(hits_one_per_locus) < len(hits)

    # If we have multiple hits and allow multiple MLST calls, then we cluster the hits by contig
    # before continuing.
    if any_multiple_hits and allow_multiple:
        hit_groups = cluster_hits_by_contig(hits)

    # If we aren't considering multiple MLST calls, then all the hits go into a single cluster.
    else:
        hit_groups = [hits_one_per_locus]

    for hit_group in hit_groups:
        call, alleles, info = \
            call_one_st(hit_group, header, check_for_truncation, max_missing, alleles_to_st,
                        required_exact_matches, info_arg, st_to_info, report_incomplete,
                        min_gene_count, unknown_group_name)
        final_call = add_to_string(final_call, call)
        final_alleles = add_to_strings(final_alleles, alleles)
        final_info = add_to_string(final_info, info)

    if spurious_hits is not None:
        spurious_hits = process_spurious_hits(spurious_hits)

    return final_call, final_alleles, final_info, spurious_hits


def call_one_st(hits, header, check_for_truncation, max_missing, alleles_to_st,
                required_exact_matches, info_arg, st_to_info, report_incomplete,
                min_gene_count, unknown_group_name):
    best_alleles, any_truncations = get_best_allele_per_locus(hits, check_for_truncation)

    best_st = []
    best_st_annotated = []

    mismatch_loci_including_snps, missing_loci = 0, 0

    for locus in header:
        if locus in best_alleles:
            allele = best_alleles[locus]

            # Remove * (inexact match) and truncation percentages
            allele_number = allele.replace('*', '')
            allele_number = re.sub(r'-\d+%', '', allele_number)

            if '*' in allele:
                mismatch_loci_including_snps += 1
            best_st.append(allele_number)
            best_st_annotated.append(allele)  # will still have character if imperfect match
        else:
            best_st.append('-')
            best_st_annotated.append('-')
            mismatch_loci_including_snps += 1
            missing_loci += 1

    # assign ST
    bst = ','.join(best_st)

    if mismatch_loci_including_snps <= max_missing:
        # only report ST if enough loci are precise matches
        if bst in alleles_to_st:
            # note may have mismatching alleles due to SNPs, this will be recorded in
            # mismatch_loci_including_snps
            bst = alleles_to_st[bst]
        else:
            # determine closest ST
            bst, _, mismatch_loci_including_snps = \
                get_closest_locus_variant(best_st, best_st_annotated, alleles_to_st)
    else:
        bst = '0'

    exact_matches = len(best_st) - mismatch_loci_including_snps
    if exact_matches < required_exact_matches:
        bst = '0'

    # pull info column
    info_final = ''
    if info_arg == 'yes' and bst in st_to_info:
        info_final = st_to_info[bst]
        if report_incomplete and missing_loci > 0:
            info_final += ' (incomplete)'
        if check_for_truncation and any_truncations:
            info_final += ' (truncated)'

    if mismatch_loci_including_snps > 0 and bst != '0':
        bst += '-' + str(mismatch_loci_including_snps) + 'LV'

    if min_gene_count is not None and unknown_group_name is not None:
        if info_final == '':
            if sum(0 if x == '-' else 1 for x in best_st_annotated) >= min_gene_count:
                info_final = unknown_group_name
                bst = '0'
            else:
                info_final = '-'

    return bst, best_st_annotated, info_final


def get_allele_and_locus(hit):
    """
    Parses the allele name and locus name from the hit's gene ID.
    """
    if '__' in hit.gene_id:  # srst2 formatted file
        gene_id_components = hit.gene_id.split('__')
        locus = gene_id_components[1]
        allele = gene_id_components[2]
    else:
        allele = hit.gene_id
        locus = hit.gene_id.split('_')[0]
    return allele, locus


def keep_only_one_hit_per_locus(hits):
    hits_per_locus = collections.defaultdict(list)
    for hit in hits:
        _, locus = get_allele_and_locus(hit)
        hits_per_locus[locus].append(hit)
    kept_hits = []
    for locus, locus_hits in hits_per_locus.items():
        locus_hits = sorted(locus_hits, key=lambda x: x.score, reverse=True)  # sort best to worst
        best_hit = locus_hits[0]
        kept_hits.append(best_hit)
    return kept_hits


def get_best_allele_per_locus(hits, check_for_truncation):
    best_scores = {}   # key = locus, value = BLAST score for best allele encountered so far
    best_alleles = {}  # key = locus, value = best allele (* if imprecise match)

    for hit in hits:
        allele, locus = get_allele_and_locus(hit)
        if hit.pcid < 100.00 or hit.alignment_length < hit.ref_length:
            allele += '*'  # inexact match
        if check_for_truncation:
            allele += truncation_check(hit)[0]
        # store best match for each one locus
        if locus in best_scores:
            if hit.score > best_scores[locus]:    # update
                best_scores[locus] = hit.score
                best_alleles[locus] = allele.split('_')[1]  # store number only
        else:  # initialise
            best_scores[locus] = hit.score
            best_alleles[locus] = allele.split('_')[1]  # store number only

    any_truncations = False
    if check_for_truncation:
        for allele in best_alleles.values():
            if allele.endswith('%'):
                any_truncations = True

    return best_alleles, any_truncations


def process_spurious_hits(hits):
    hit_strings = []
    for hit in hits:
        allele, locus = get_allele_and_locus(hit)
        if hit.pcid < 100.00 or hit.alignment_length < hit.ref_length:
            allele += '*'  # inexact match
        allele += truncation_check(hit)[0]
        hit_strings.append(allele)
    return hit_strings


def load_st_database(database, info_arg):
    st_names = []
    alleles_to_st = {}  # key = concatenated string of alleles, value = st
    st_to_info = {}  # key = st, value = info relating to this ST, eg clonal group
    header = []
    with open(database, 'r') as f:
        for line in f:
            fields = line.rstrip().split('\t')
            if len(header) == 0:
                header = fields
                header.pop(0)  # remove st label
                if info_arg == 'yes':
                    header.pop()  # remove info label
            else:
                st = fields.pop(0)
                if info_arg == 'yes':
                    info = fields.pop()
                else:
                    info = ''
                st_names.append(st)
                alleles_to_st[','.join(fields)] = st
                if info_arg == 'yes':
                    st_to_info[st] = info
    return st_names, alleles_to_st, st_to_info, header


def get_closest_locus_variant(query, annotated_query, sts):
    annotated_query = list(annotated_query)  # copy the list so we don't change the original
    closest = []
    closest_alleles = {}   # key = st, value = list
    min_dist = len(query)  # number mismatching loci, ignoring SNPs

    for index, item in enumerate(query):
        if item == '-':
            query[index] = '0'

    # get distance from closest ST, ignoring SNPs
    for st in sts:
        d = sum(map(lambda x, y: bool(int(x)-int(y)), st.split(','), query))
        if d == min_dist:
            closest.append(int(sts[st]))
            closest_alleles[sts[st]] = st
        elif d < min_dist:
            # reset
            closest = [int(sts[st])]
            closest_alleles[sts[st]] = st
            min_dist = d  # distance from closest ST, ignoring SNPs

    closest_st = str(min(closest))

    for index, item in enumerate(annotated_query):
        annotated_query[index] = re.sub(r'-\d+%', '', item)
        if item == '-' or '*' in item:
            annotated_query[index] = '0'

    # get distance from closest ST, including SNPs
    min_dist_incl_snps = sum(map(lambda x, y: bool(int(x)-int(y)),
                                 closest_alleles[closest_st].split(','), annotated_query))

    return closest_st, min_dist, min_dist_incl_snps


def add_to_string(existing_str, new_str):
    if existing_str == '':
        return new_str
    elif new_str == '':
        return existing_str
    else:
        return existing_str + ',' + new_str


def add_to_strings(existing_strs, new_strs):
    assert len(existing_strs) == len(new_strs)
    return [add_to_string(existing_str, new_str)
            for existing_str, new_str in zip(existing_strs, new_strs)]


def cluster_hits_by_contig(hits):
    hits_by_contig = collections.defaultdict(list)
    for h in hits:
        hits_by_contig[h.contig_name].append(h)
    return [x for x in hits_by_contig.values()]



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_iro_mlst_results


class TestIro(unittest.TestCase):
    """
    Tests salmochelin ST calls.
    """

    def setUp(self):
        self.data_dir = 'test/test_iro/data'
        Args = collections.namedtuple('Args', ['min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)

    def test_iro_random(self):
        """
        This test has just random sequence and should give no iro call.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_random.fasta', self.args)
        self.assertEqual(results['iroB'], '-')
        self.assertEqual(results['iroC'], '-')
        self.assertEqual(results['iroD'], '-')
        self.assertEqual(results['iroN'], '-')
        self.assertEqual(results['Salmochelin'], '-')
        self.assertEqual(results['SmST'], '0')

    def test_iro_exact_33(self):
        """
        This test is an exact match for SmST33.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_1.fasta', self.args)
        self.assertEqual(results['iroB'], '22-29%')
        self.assertEqual(results['iroC'], '33-4%')
        self.assertEqual(results['iroD'], '12')
        self.assertEqual(results['iroN'], '15')
        self.assertEqual(results['Salmochelin'], 'iro 4 (truncated)')
        self.assertEqual(results['SmST'], '33')

    def test_iro_inexact(self):
        """
        This test is an inexact match for SmST33. There are single base changes in iroC and in
        iroN.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_2.fasta', self.args)
        self.assertEqual(results['iroB'], '22-29%')
        self.assertEqual(results['iroC'], '33*-4%')
        self.assertEqual(results['iroD'], '12')
        self.assertEqual(results['iroN'], '15*')
        self.assertEqual(results['Salmochelin'], 'iro 4 (truncated)')
        self.assertEqual(results['SmST'], '33-2LV')

    def test_iro_novel(self):
        """
        This test is an exact match for alleles, but an unknown combination.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_3.fasta', self.args)
        self.assertEqual(results['iroB'], '11')
        self.assertEqual(results['iroC'], '1')
        self.assertEqual(results['iroD'], '18')
        self.assertEqual(results['iroN'], '5')
        self.assertEqual(results['Salmochelin'], 'iro unknown')
        self.assertEqual(results['SmST'], '0')

    def test_iro_incomplete(self):
        """
        This test is an exact match for only one allele.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_4.fasta', self.args)
        self.assertEqual(results['iroB'], '-')
        self.assertEqual(results['iroC'], '1')
        self.assertEqual(results['iroD'], '-')
        self.assertEqual(results['iroN'], '-')
        self.assertEqual(results['Salmochelin'], '-')
        self.assertEqual(results['SmST'], '0')

    def test_iro_exact_1(self):
        """
        This test is an exact match for SmST1.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_5.fasta', self.args)
        self.assertEqual(results['iroB'], '1')
        self.assertEqual(results['iroC'], '1')
        self.assertEqual(results['iroD'], '1')
        self.assertEqual(results['iroN'], '1')
        self.assertEqual(results['Salmochelin'], 'iro 1')
        self.assertEqual(results['SmST'], '1')

    def test_iro_missing_1(self):
        """
        This test is a match for SmST1, except that it's missing iroN.
        """
        results = get_iro_mlst_results(self.data_dir, 'test/test_iro/test_iro_6.fasta', self.args)
        self.assertEqual(results['iroB'], '1')
        self.assertEqual(results['iroC'], '1')
        self.assertEqual(results['iroD'], '1')
        self.assertEqual(results['iroN'], '-')
        self.assertEqual(results['Salmochelin'], 'iro 1 (incomplete)')
        self.assertEqual(results['SmST'], '1-1LV')



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_output_headers, get_resistance_results
from kleborate.resBLAST import is_exact_aa_match


class TestResAlleles(unittest.TestCase):
    """
    Tests calling of resistance via alleles.
    """

    def setUp(self):
        self.data_dir = 'test/test_res_alleles/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def test_res_01(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/01.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'ABC-1')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_02(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/02.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'ABC-2')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_03(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/03.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], 'ABC-3')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_04(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/04.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], 'ABC-4')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_05(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/05.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertTrue(results['Bla_acquired'] == 'ABC-1^' or results['Bla_acquired'] == 'ABC-2^')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_06(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/06.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertTrue(results['Bla_ESBL_acquired'] == 'ABC-3^' or
                        results['Bla_ESBL_acquired'] == 'ABC-4^')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_07(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/07.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'ABC-2*')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_08(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/08.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'ABC-1;ABC-2')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_09(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/09.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertTrue(results['Bla_acquired'] == 'ABC-1^' or results['Bla_acquired'] == 'ABC-2^')
        self.assertTrue(results['Bla_ESBL_acquired'] == 'ABC-3^' or
                        results['Bla_ESBL_acquired'] == 'ABC-4^')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_10(self):
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/10.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], 'ABC-1;ABC-2')
        self.assertEqual(results['Bla_ESBL_acquired'], 'ABC-3*;ABC-4')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_11(self):
        """
        This test has the ABC-1 gene but with a stop codon in the middle.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/11.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['truncated_resistance_hits'], 'ABC-1*-50%')

    def test_res_12(self):
        """
        This test has the ABC-1 gene but missing the first (start) codon.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/12.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['truncated_resistance_hits'], 'ABC-1?-0%')

    def test_res_13(self):
        """
        This test has the ABC-1 gene but at only 50% coverage, so it goes to spurious hits.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/13.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Tet_acquired'], '-')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], 'ABC-1?-50%')

    def test_res_14(self):
        """
        I added this test to catch a weird BLAST-related bug where an exact amino acid match for
        SHV-51 was present but not being found.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/14.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-51^')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_15(self):
        """
        Same as the previous test, but with the nucleotide sequence reverse-complemented.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/15.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'SHV-51^')
        self.assertEqual(results['Bla_acquired'], '-')
        self.assertEqual(results['Bla_ESBL_acquired'], '-')
        self.assertEqual(results['spurious_resistance_hits'], '-')

    def test_res_16(self):
        """
        I added this test to catch a weird BLAST-related bug where an exact amino acid match for
        sul2 was present but not being found.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/16.fasta', self.args,
                                         self.res_headers, True)
        print(results)
        self.assertEqual(results['Sul_acquired'], 'sul2^')

    def test_res_17(self):
        """
        Same as the previous test, but with the nucleotide sequence reverse-complemented.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/17.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Sul_acquired'], 'sul2^')

    def test_res_18(self):
        """
        This is the uncomplicated case for a OKP-B-6^ result (later tests cover more challenging
        cases).
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/18.fasta', self.args,
                                         self.res_headers, True)
        print(results)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_res_19(self):
        """
        Same as the previous test, but with the nucleotide sequence reverse-complemented.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/19.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_res_20(self):
        """
        I added this test to catch a bug where there is an exact amino acid match but inexact
        nucleotide match, and it's the very last base that differs (alternate stop codon).
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/20.fasta', self.args,
                                         self.res_headers, True)
        print(results)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_res_21(self):
        """
        Same as the previous test, but with the nucleotide sequence reverse-complemented.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/21.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_res_22(self):
        """
        I added this test to catch a bug where there is an exact amino acid match but inexact
        nucleotide match, and it's the very first base that differs (alternate start codon).
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/22.fasta', self.args,
                                         self.res_headers, True)
        print(results)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_res_23(self):
        """
        Same as the previous test, but with the nucleotide sequence reverse-complemented.
        """
        results = get_resistance_results(self.data_dir, 'test/test_res_alleles/23.fasta', self.args,
                                         self.res_headers, True)
        self.assertEqual(results['Bla_chr'], 'OKP-B-6^')

    def test_exact_aa_match_1(self):
        """
        Simple case: full protein sequence ending in a stop codon.
          seq_1 -> MNKSLIIFGI*
          seq_2 -> MNKSLIIFGI*
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCGGCATCTAA'
        seq_2 = 'ATGAATAAATCGCTAATCATTTTCGGCATCTAA'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_2(self):
        """
        These two seqs have a stop codon in the middle but still match exactly.
          seq_1 -> MNKSLIIF*GI
          seq_2 -> MNKSLIIF*GI
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCTAAGGCATC'
        seq_2 = 'ATGAACAAGTCGCTCATCATTTTCTAAGGCATC'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_3(self):
        """
        These two seqs have a stop codon in the middle and mismatch after the stop.
          seq_1 -> MNKSLIIF*GI
          seq_2 -> MNKSLIIF*TL
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCTAAGGCATC'
        seq_2 = 'ATGAACAAGTCGCTCATCATTTTCTAAACCTTA'
        self.assertFalse(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_4(self):
        """
        These two seqs have a stop codon in the middle and mismatch after the stop.
          seq_1 -> MNK*NYRTDYD
          seq_2 -> MNK*VD*LAGP
        """
        seq_1 = 'ATGAATAAATAAAACTATCGTACGGACTACGAC'
        seq_2 = 'ATGAACAAGTAAGTCGACTGACTAGCTGGACCC'
        self.assertFalse(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_5(self):
        """
        These two seqs are non-multiple-of-three in length and match in their remainder.
          seq_1 -> MNKSLIIFG (+AT)
          seq_2 -> MNKSLIIFG (+AT)
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCGGCAT'
        seq_2 = 'ATGAACAAGTCGCTCATCATTTTCGGGAT'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_6(self):
        """
        These two seqs are non-multiple-of-three in length and match in their remainder.
          seq_1 -> MNKSLIIFG (+AT)
          seq_2 -> MNKSLIIFG (+A)
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCGGCAT'
        seq_2 = 'ATGAACAAGTCGCTCATCATTTTCGGGA'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_7(self):
        """
        These two seqs are non-multiple-of-three in length and don't match in their remainder.
          seq_1 -> MNKSLIIFG (+AT)
          seq_2 -> MNKSLIIFG (+CC)
        """
        seq_1 = 'ATGAATAAATCGCTAATCATTTTCGGCAT'
        seq_2 = 'ATGAACAAGTCGCTCATCATTTTCGGGCC'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))

    def test_exact_aa_match_8(self):
        """
        The first seq's protein sequence contains the second seq's protein sequence.
          seq_1 -> DTMNKSLGI*RS
          seq_2 -> MNKSLGI*
        """
        seq_1 = 'GACACGATGAATAAATCGCTAGGCATCTAACGATCA'
        seq_2 = 'ATGAACAAGTCGCTCGGGATCTAA'
        self.assertTrue(is_exact_aa_match(seq_1, seq_2))



#!/usr/bin/env python3
"""
Copyright 2018 Kat Holt
Copyright 2018 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import sys
import argparse
import collections
import pandas as pd
from Bio import Phylo


def get_arguments():
    parser = argparse.ArgumentParser(description='A script for converting Kleborate output into a'
                                                 'format compatible with Microreact')
    required_args = parser.add_argument_group('Required arguments')
    required_args.add_argument('--kleborate_in', type=str, required=True,
                               help='Kleborate tab-delimited results file')
    required_args.add_argument('--tree_in', type=str, required=True,
                               help='Phylogenetic tree')
    required_args.add_argument('--csv_out', type=str, required=True,
                               help='Kleborate results in Microreact format')
    required_args.add_argument('--tree_out', type=str, required=True,
                               help='Tree in Microreact format')

    return parser.parse_args()


def main():
    args = get_arguments()
    name_subs = name_substitution(args.kleborate_in)
    check_for_unique_names(name_subs)
    save_tree_with_new_names(args.tree_in, args.tree_out, name_subs)
    autocolour_columns = get_autocolour_columns(args.kleborate_in)

    csv_lines = []
    with open(args.kleborate_in, 'rt') as kleborate_results:
        original_header, new_header = None, None
        for line in kleborate_results:
            line = line.rstrip('\n')
            if original_header is None:
                original_header = line.split('\t')
                new_header = get_new_header(original_header, autocolour_columns)
                line_parts = new_header
            else:
                line_parts = get_data(line, name_subs, original_header, new_header)
            csv_lines.append((','.join(line_parts)))

    print()
    print('Writing Microreact table to: {}'.format(args.csv_out))
    with open(args.csv_out, 'wt') as output_csv:
        for line in csv_lines:
            output_csv.write(line)
            output_csv.write('\n')

    print()




def get_autocolour_columns(kleborate_in):
    autocolour_columns = []
    table = pd.read_table(kleborate_in)
    for col_name in ['species', 'ST', 'YbST', 'CbST', 'AbST', 'SmST', 'wzi', 'K_locus',
                     'O_locus']:
        try:
            if len(set(table[col_name])) > 1:
                autocolour_columns.append(col_name)
        except KeyError:
            pass
    print()
    print('Using "__autocolour" on the following columns:')
    print('   ', ', '.join(autocolour_columns))
    return set(autocolour_columns)


def get_new_header(original_header, autocolour_columns):
    original_header[0] = 'id'  # Change 'strain' to 'id' for Microreact.
    for autocolour_column in autocolour_columns:
        i = find_column_index(original_header, autocolour_column)
        original_header[i] = autocolour_column + '__autocolour'

    header = list(original_header)
    for col in ['virulence_score', 'resistance_score', 'num_resistance_classes',
                'num_resistance_genes', 'Yersiniabactin', 'Colibactin', 'Aerobactin',
                'Salmochelin', 'rmpA', 'rmpA2']:
        header.insert(find_column_index(header, col) + 1, col + '__colour')

    for res in ['AGly_acquired', 'Col_acquired', 'Fcyn_acquired', 'Flq_acquired', 'Gly_acquired', 'MLS_acquired', 'Phe_acquired', 'Rif_acquired', 'Sul_acquired', 'Tet_acquired',
                'Tgc_acquired', 'Tmt_acquired', 'Bla_acquired', 'Bla_inhR_acquired', 'Bla_ESBL_acquired', 'Bla_ESBL_inhR_acquired', 'Bla_Carb_acquired', 'Bla_chr', 'SHV_mutations', 
                'Col_mutations', 'Flq_mutations', 'Omp_mutations']:
        header.insert(find_column_index(header, res) + 1, res + '__colour')
        header.remove(res)

    return header


def get_data(line, name_subs, original_header, new_header):
    line = line.replace(',', ';')
    line_parts = line.split('\t')

    line_parts[0] = name_subs[line_parts[0]]

    original_data = dict(zip(original_header, line_parts))
    new_data = {h: '' for h in new_header}

    for label, value in original_data.items():
        new_data[label] = value

    vir_score = int(original_data['virulence_score'])
    res_score = int(original_data['resistance_score'])
    res_classes = int(original_data['num_resistance_classes'])
    res_genes = int(original_data['num_resistance_genes'])

    new_data['virulence_score__colour'] = get_vir_score_colour(vir_score)
    new_data['resistance_score__colour'] = get_res_score_colour(res_score)
    new_data['num_resistance_classes__colour'] = get_res_classes_colour(res_classes)
    new_data['num_resistance_genes__colour'] = get_res_genes_colour(res_genes)
#    new_data['Yersiniabactin__colour'] = get_vir_lineage_colour(original_data['Yersiniabactin'])
    new_data['Colibactin__colour'] = get_vir_lineage_colour(original_data['Colibactin'])
    new_data['Aerobactin__colour'] = get_vir_lineage_colour(original_data['Aerobactin'])
    new_data['Salmochelin__colour'] = get_vir_lineage_colour(original_data['Salmochelin'])
    new_data['rmpA__colour'] = get_rmpA_colour(original_data['rmpA'])
    new_data['rmpA2__colour'] = get_rmpA2_colour(original_data['rmpA2'])
    for res_class in ['AGly_acquired', 'Col_acquired', 'Fcyn_acquired', 'Flq_acquired', 'Gly_acquired', 'MLS_acquired', 'Phe_acquired', 'Rif_acquired', 'Sul_acquired', 'Tet_acquired',
                      'Tgc_acquired', 'Tmt_acquired', 'Bla_acquired', 'Bla_inhR_acquired', 'Bla_ESBL_acquired', 'Bla_ESBL_inhR_acquired', 'Bla_Carb_acquired', 'Bla_chr', 'SHV_mutations', 
                      'Col_mutations', 'Flq_mutations', 'Omp_mutations']:
        new_data[res_class + '__colour'] = get_res_class_colour(original_data[res_class])

    return [new_data[h] for h in new_header]


def name_substitution(kleborate_in):
    name_subs = {}
    with open(kleborate_in, 'rt') as kleborate_results:
        header = None
        for line in kleborate_results:
            if header is None:
                header = line.split('\t')
                if header[0] != 'strain':
                    sys.exit('Error: first column is not "strain" - is this Kleborate output?')
            else:
                line_parts = line.split('\t')
                if len(line_parts) != len(header):
                    sys.exit('Error: inconsistent number of columns')
                old_name = line_parts[0]
                if old_name in name_subs:
                    sys.exit('Error: duplicate sample ID: ' + old_name)
                new_name = old_name.replace('.', '_')
                new_name = new_name.replace(',', '_')
                new_name = new_name.replace("'", '_')
                new_name = new_name.replace('"', '_')
                name_subs[old_name] = new_name
    return name_subs


def check_for_unique_names(name_subs):
    names = list(name_subs.values())
    duplicate_names = [item for item, count in collections.Counter(names).items() if count > 1]
    if duplicate_names:
        sys.exit('Error: duplicate sample IDs: ' + ', '.join(duplicate_names))


def save_tree_with_new_names(tree_in, tree_out, name_subs):
    print()
    print('Writing Microreact tree to: {}'.format(tree_out))
    tree_format = None
    for try_tree_format in ['newick', 'nexus', 'nexml', 'phyloxml', 'cdao']:
        try:
            Phylo.read(tree_in, try_tree_format)
            tree_format = try_tree_format
            break
        except ValueError:
            pass
    if tree_format is None:
        sys.exit('Error: could not read input tree')

    tree = Phylo.read(tree_in, tree_format)
    for node in tree.get_terminals():
        name = str(node.name)
        try:
            node.name = name_subs[name]
        except IndexError:
            sys.exit('Error: sample name in tree not in Kleborate data: ' + name)
    Phylo.write(tree, tree_out, 'newick')


def scale_num(start, end, progress):
    return int(round(start * (1.0 - progress) + end * progress))


def colour_range(start, end, count):
    start, end = start.lower(), end.lower()
    if start.startswith('#'):
        start = start[1:]
    if end.startswith('#'):
        end = end[1:]
    start_r, start_g, start_b = int(start[0:2], 16), int(start[2:4], 16), int(start[4:6], 16)
    end_r, end_g, end_b = int(end[0:2], 16), int(end[2:4], 16), int(end[4:6], 16)
    colours = []
    for i in range(count):
        progress = i / (count - 1)
        r, g, b = scale_num(start_r, end_r, progress), scale_num(start_g, end_g, progress), \
            scale_num(start_b, end_b, progress)
        hex_colour = '"#' + ('0x%X' % r)[2:] + ('0x%X' % g)[2:] + ('0x%X' % b)[2:] + '"'
        colours.append(hex_colour)
    return colours


def find_column_index(header, col_name):
    try:
        return header.index(col_name)
    except ValueError:
        sys.exit('Error: could not find ' + col_name + ' column in Kleborate')


def get_vir_score_colour(vir_score):
    try:
        return ['#DEEBF7', '#9ECAE1', '#6BAED6', '#4292C6', '#2171B5', '#08306B'][vir_score]
    except IndexError:
        return '#BFBFBF'


def get_res_score_colour(res_score):
    try:
        return ['#FCBBA1', '#FC9272', '#FB6A4A', '#BE413D'][res_score]
    except IndexError:
        return '#BFBFBF'


def get_res_classes_colour(res_classes):
    try:
        return colour_range('#FCBBA1', '#BE413D', 11)[res_classes]
    except IndexError:
        return '#BE413D'


def get_res_genes_colour(res_genes):
    try:
        return colour_range('#FCBBA1', '#BE413D', 21)[res_genes]
    except IndexError:
        return '#BE413D'


def get_species_colour(species):
    try:
        return {'Klebsiella pneumoniae': '#875F9A',
                'Klebsiella variicola subsp. variicola': '#8CBDB2',
                'Klebsiella quasivariicola': '#F0B663',
                'Klebsiella quasipneumoniae subsp. quasipneumoniae': '#ED6060',
                'Klebsiella quasipneumoniae subsp. similipneumoniae': '#EDA483'}[species]
    except IndexError:
        return '#BFBFBF'


def get_vir_lineage_colour(vir_lineage):
    vir_lineage_colours = {'ybt 1': '#B27F91', 'ybt 2': '#CDA12C', 'ybt 3': '#56A354',
                           'ybt 4': '#F28FA2', 'ybt 5': '#DB7723', 'ybt 6': '#93539D',
                           'ybt 7': '#3A85A8', 'ybt 8': '#7B75CC', 'ybt 9': '#D9C5EF',
                           'ybt 10': '#449D72', 'ybt 11': '#EBD930', 'ybt 12': '#6AA3C6',
                           'ybt 13': '#A39F93', 'ybt 14': '#93539D', 'ybt 15': '#EDC59A',
                           'ybt 16': '#840639', 'ybt 17': '#E25065', 'clb 1': '#99BBE0',
                           'clb 2A': '#5972AF', 'clb 2B': '#242F69', 'clb 3': '#242F69',
                           'iro 1': '#B6D5EF', 'iro 2': '#DEC4E8', 'iro 3': '#E29771',
                           'iro 4': '#A4A4EA', 'iro 5': '#E0AAAA', 'iuc 1': '#B6D5EF',
                           'iuc 2': '#DEC4E8', 'iuc 2A': '#D8ABDD', 'iuc 3': '#C3EADB',
                           'iuc 4': '#9ACCBC', 'iuc 5': '#E0AAAA'}
    vir_lineage = vir_lineage.split(';')[0]
    if vir_lineage in vir_lineage_colours:
        return vir_lineage_colours[vir_lineage]
    elif vir_lineage == '-':
        return '#FFFFFF'
    else:
        return '#BFBFBF'


def get_rmpA_colour(rmpA):
    return '#FFFFFF' if rmpA == '-' else '#08306B'
    
def get_rmpA2_colour(rmpA2):
    return '#FFFFFF' if rmpA2 == '-' else '#08306B'


def get_res_class_colour(res_class):
    return '#FFFFFF' if res_class == '-' else '#BE413D'


if __name__ == '__main__':
    main()
